#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, Mock, patch

import pytest
from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.speech_v1 import RecognitionAudio, RecognitionConfig, RecognizeResponse

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.operators.speech_to_text import CloudSpeechToTextRecognizeSpeechOperator

PROJECT_ID = "project-id"
GCP_CONN_ID = "gcp-conn-id"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
CONFIG = RecognitionConfig({"encoding": "LINEAR16"})
AUDIO = RecognitionAudio({"uri": "gs://bucket/object"})


class TestCloudSpeechToTextRecognizeSpeechOperator:
    @patch("airflow.providers.google.cloud.operators.speech_to_text.CloudSpeechToTextHook")
    def test_recognize_speech_green_path(self, mock_hook):
        mock_hook.return_value.recognize_speech.return_value = RecognizeResponse()

        CloudSpeechToTextRecognizeSpeechOperator(
            project_id=PROJECT_ID,
            gcp_conn_id=GCP_CONN_ID,
            config=CONFIG,
            audio=AUDIO,
            task_id="id",
            impersonation_chain=IMPERSONATION_CHAIN,
        ).execute(context=MagicMock())

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.recognize_speech.assert_called_once_with(
            config=CONFIG, audio=AUDIO, retry=DEFAULT, timeout=None
        )

    @patch("airflow.providers.google.cloud.operators.speech_to_text.CloudSpeechToTextHook")
    def test_missing_config(self, mock_hook):
        mock_hook.return_value.recognize_speech.return_value = True

        with pytest.raises((TypeError, AirflowException), match="missing keyword argument 'config'"):
            CloudSpeechToTextRecognizeSpeechOperator(
                project_id=PROJECT_ID, gcp_conn_id=GCP_CONN_ID, audio=AUDIO, task_id="id"
            ).execute(context={"task_instance": Mock()})
        mock_hook.assert_not_called()

    @patch("airflow.providers.google.cloud.operators.speech_to_text.CloudSpeechToTextHook")
    def test_missing_audio(self, mock_hook):
        mock_hook.return_value.recognize_speech.return_value = True

        with pytest.raises((TypeError, AirflowException), match="missing keyword argument 'audio'"):
            CloudSpeechToTextRecognizeSpeechOperator(
                project_id=PROJECT_ID, gcp_conn_id=GCP_CONN_ID, config=CONFIG, task_id="id"
            ).execute(context={"task_instance": Mock()})
        mock_hook.assert_not_called()

    @patch("airflow.providers.google.cloud.operators.speech_to_text.FileDetailsLink.persist")
    @patch("airflow.providers.google.cloud.operators.speech_to_text.CloudSpeechToTextHook")
    def test_no_audio_uri(self, mock_hook, mock_file_link):
        mock_hook.return_value.recognize_speech.return_value = RecognizeResponse()
        AUDIO_NO_URI = RecognitionAudio({"content": b"set content data instead of uri"})

        op = CloudSpeechToTextRecognizeSpeechOperator(
            project_id=PROJECT_ID,
            gcp_conn_id=GCP_CONN_ID,
            config=CONFIG,
            audio=AUDIO_NO_URI,
            task_id="id",
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context=MagicMock())

        mock_hook.return_value.recognize_speech.assert_called_once_with(
            config=CONFIG, audio=AUDIO_NO_URI, retry=DEFAULT, timeout=None
        )
        assert op.audio.uri == ""
        mock_file_link.assert_not_called()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


"""Example Airflow DAG for Google Vertex AI service testing Custom Jobs operations."""

from __future__ import annotations

import os
from datetime import datetime

from google.cloud.aiplatform import schema
from google.protobuf.json_format import ParseDict
from google.protobuf.struct_pb2 import Value

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import (
    GCSCreateBucketOperator,
    GCSDeleteBucketOperator,
    GCSSynchronizeBucketsOperator,
)
from airflow.providers.google.cloud.operators.vertex_ai.custom_job import (
    CreateCustomPythonPackageTrainingJobOperator,
    DeleteCustomTrainingJobOperator,
)
from airflow.providers.google.cloud.operators.vertex_ai.dataset import (
    CreateDatasetOperator,
    DeleteDatasetOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
DAG_ID = "vertex_ai_custom_job_operations"
REGION = "us-central1"
PACKAGE_DISPLAY_NAME = f"train-housing-py-package-{ENV_ID}"
MODEL_DISPLAY_NAME = f"py-package-housing-model-{ENV_ID}"

RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"
CUSTOM_PYTHON_GCS_BUCKET_NAME = f"bucket_python_{DAG_ID}_{ENV_ID}".replace("_", "-")

DATA_SAMPLE_GCS_OBJECT_NAME = "vertex-ai/california_housing_train.csv"


def TABULAR_DATASET(bucket_name):
    return {
        "display_name": f"tabular-dataset-{ENV_ID}",
        "metadata_schema_uri": schema.dataset.metadata.tabular,
        "metadata": ParseDict(
            {"input_config": {"gcs_source": {"uri": [f"gs://{bucket_name}/{DATA_SAMPLE_GCS_OBJECT_NAME}"]}}},
            Value(),
        ),
    }


CONTAINER_URI = "us-docker.pkg.dev/vertex-ai/training/tf-cpu.2-2:latest"
MODEL_SERVING_CONTAINER_URI = "us-docker.pkg.dev/vertex-ai/prediction/tf2-cpu.2-2:latest"
REPLICA_COUNT = 1
MACHINE_TYPE = "n1-standard-4"
ACCELERATOR_TYPE = "ACCELERATOR_TYPE_UNSPECIFIED"
ACCELERATOR_COUNT = 0
TRAINING_FRACTION_SPLIT = 0.7
TEST_FRACTION_SPLIT = 0.15
VALIDATION_FRACTION_SPLIT = 0.15

PYTHON_PACKAGE_GCS_URI = f"gs://{CUSTOM_PYTHON_GCS_BUCKET_NAME}/vertex-ai/custom_trainer_script-0.1.tar"
PYTHON_MODULE_NAME = "aiplatform_custom_trainer_script.task"


with DAG(
    f"{DAG_ID}_python_package",
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "vertex_ai", "custom_job"],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=CUSTOM_PYTHON_GCS_BUCKET_NAME,
        storage_class="REGIONAL",
        location=REGION,
    )

    move_data_files = GCSSynchronizeBucketsOperator(
        task_id="move_files_to_bucket",
        source_bucket=RESOURCE_DATA_BUCKET,
        source_object="vertex-ai/california-housing-data",
        destination_bucket=CUSTOM_PYTHON_GCS_BUCKET_NAME,
        destination_object="vertex-ai",
        recursive=True,
    )

    create_tabular_dataset = CreateDatasetOperator(
        task_id="tabular_dataset",
        dataset=TABULAR_DATASET(CUSTOM_PYTHON_GCS_BUCKET_NAME),
        region=REGION,
        project_id=PROJECT_ID,
    )
    tabular_dataset_id = create_tabular_dataset.output["dataset_id"]

    # [START how_to_cloud_vertex_ai_create_custom_python_package_training_job_operator]
    create_custom_python_package_training_job = CreateCustomPythonPackageTrainingJobOperator(
        task_id="python_package_task",
        staging_bucket=f"gs://{CUSTOM_PYTHON_GCS_BUCKET_NAME}",
        display_name=PACKAGE_DISPLAY_NAME,
        python_package_gcs_uri=PYTHON_PACKAGE_GCS_URI,
        python_module_name=PYTHON_MODULE_NAME,
        container_uri=CONTAINER_URI,
        model_serving_container_image_uri=MODEL_SERVING_CONTAINER_URI,
        # run params
        dataset_id=tabular_dataset_id,
        model_display_name=MODEL_DISPLAY_NAME,
        replica_count=REPLICA_COUNT,
        machine_type=MACHINE_TYPE,
        accelerator_type=ACCELERATOR_TYPE,
        accelerator_count=ACCELERATOR_COUNT,
        training_fraction_split=TRAINING_FRACTION_SPLIT,
        validation_fraction_split=VALIDATION_FRACTION_SPLIT,
        test_fraction_split=TEST_FRACTION_SPLIT,
        region=REGION,
        project_id=PROJECT_ID,
    )
    # [END how_to_cloud_vertex_ai_create_custom_python_package_training_job_operator]

    # [START how_to_cloud_vertex_ai_create_custom_python_package_training_job_operator_deferrable]
    create_custom_python_package_training_job_deferrable = CreateCustomPythonPackageTrainingJobOperator(
        task_id="python_package_task_deferrable",
        staging_bucket=f"gs://{CUSTOM_PYTHON_GCS_BUCKET_NAME}",
        display_name=f"{PACKAGE_DISPLAY_NAME}-def",
        python_package_gcs_uri=PYTHON_PACKAGE_GCS_URI,
        python_module_name=PYTHON_MODULE_NAME,
        container_uri=CONTAINER_URI,
        model_serving_container_image_uri=MODEL_SERVING_CONTAINER_URI,
        # run params
        dataset_id=tabular_dataset_id,
        model_display_name=f"{MODEL_DISPLAY_NAME}-def",
        replica_count=REPLICA_COUNT,
        machine_type=MACHINE_TYPE,
        accelerator_type=ACCELERATOR_TYPE,
        accelerator_count=ACCELERATOR_COUNT,
        training_fraction_split=TRAINING_FRACTION_SPLIT,
        validation_fraction_split=VALIDATION_FRACTION_SPLIT,
        test_fraction_split=TEST_FRACTION_SPLIT,
        region=REGION,
        project_id=PROJECT_ID,
        deferrable=True,
    )
    # [END how_to_cloud_vertex_ai_create_custom_python_package_training_job_operator_deferrable]

    delete_custom_training_job = DeleteCustomTrainingJobOperator(
        task_id="delete_custom_training_job",
        training_pipeline_id="{{ task_instance.xcom_pull(task_ids='python_package_task', "
        "key='training_id') }}",
        custom_job_id="{{ task_instance.xcom_pull(task_ids='python_package_task', key='custom_job_id') }}",
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_custom_training_job_deferrable = DeleteCustomTrainingJobOperator(
        task_id="delete_custom_training_job_deferrable",
        training_pipeline_id="{{ task_instance.xcom_pull(task_ids='python_package_task_deferrable', "
        "key='training_id') }}",
        custom_job_id="{{ task_instance.xcom_pull(task_ids='python_package_task_deferrable', key='custom_job_id') }}",
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_tabular_dataset = DeleteDatasetOperator(
        task_id="delete_tabular_dataset",
        dataset_id=tabular_dataset_id,
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=CUSTOM_PYTHON_GCS_BUCKET_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        create_bucket
        >> move_data_files
        >> create_tabular_dataset
        # TEST BODY
        >> [create_custom_python_package_training_job, create_custom_python_package_training_job_deferrable]
        # TEST TEARDOWN
        >> delete_custom_training_job
        >> delete_custom_training_job_deferrable
        >> delete_tabular_dataset
        >> delete_bucket
    )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from airflow.providers.common.compat.sdk import BaseOperatorLink, XCom

if TYPE_CHECKING:
    from airflow.models.taskinstancekey import TaskInstanceKey
    from airflow.providers.common.compat.sdk import Context
    from airflow.providers.google.version_compat import BaseOperator


BASE_LINK = "https://analytics.google.com/analytics/web/"


class GoogleAnalyticsBaseLink(BaseOperatorLink):
    """
    Base class for Google Analytics links.

    :meta private:
    """

    name: ClassVar[str]
    key: ClassVar[str]
    format_str: ClassVar[str]

    def get_link(self, operator: BaseOperator, *, ti_key: TaskInstanceKey) -> str:
        if conf := XCom.get_value(key=self.key, ti_key=ti_key):
            res = BASE_LINK + "#/" + self.format_str.format(**conf)
            return res
        return ""


class GoogleAnalyticsPropertyLink(GoogleAnalyticsBaseLink):
    """Helper class for constructing Google Analytics Property Link."""

    name = "Data Analytics Property"
    key = "data_analytics_property"
    format_str = "p{property_id}/"

    @staticmethod
    def persist(
        context: Context,
        property_id: str,
    ):
        context["task_instance"].xcom_push(
            key=GoogleAnalyticsPropertyLink.key,
            value={"property_id": property_id},
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock

from airflow.providers.google.cloud.transfers.bigquery_to_mysql import BigQueryToMySqlOperator

TASK_ID = "test-bq-create-table-operator"
TEST_DATASET = "test-dataset"
TEST_TABLE_ID = "test-table-id"
TEST_DAG_ID = "test-bigquery-operators"
TEST_PROJECT = "test-project"


def _make_bq_table(schema_names: list[str]):
    class TableObj:
        def __init__(self, schema):
            self.schema = []
            for n in schema:
                field = MagicMock()
                field.name = n
                self.schema.append(field)
            self.description = "table description"
            self.external_data_configuration = None
            self.labels = {}
            self.num_rows = 0
            self.num_bytes = 0
            self.table_type = "TABLE"

    return TableObj(schema_names)


class TestBigQueryToMySqlOperator:
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_sql.BigQueryHook")
    def test_execute_good_request_to_bq(self, mock_hook):
        destination_table = "table"
        operator = BigQueryToMySqlOperator(
            task_id=TASK_ID,
            dataset_table=f"{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name=destination_table,
            replace=False,
        )

        operator.execute(None)
        mock_hook.return_value.list_rows.assert_called_once_with(
            dataset_id=TEST_DATASET,
            table_id=TEST_TABLE_ID,
            max_results=1000,
            selected_fields=None,
            start_index=0,
        )

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_sql.BigQueryHook")
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_mysql.MySqlHook")
    def test_get_openlineage_facets_on_complete_no_selected_fields(self, mock_mysql_hook, mock_bq_hook):
        mock_bq_client = MagicMock()
        mock_bq_client.get_table.return_value = _make_bq_table(["id", "name", "value"])
        mock_bq_hook.get_client.return_value = mock_bq_client
        mock_bq_hook.return_value = mock_bq_hook

        db_info = MagicMock(scheme="mysql", authority="localhost:3306", database="mydb")
        mock_mysql_hook.get_openlineage_database_info.return_value = db_info
        mock_mysql_hook.return_value = mock_mysql_hook

        op = BigQueryToMySqlOperator(
            task_id=TASK_ID,
            dataset_table=f"{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name="destination",
            selected_fields=None,
            database="mydb",
        )
        op.bigquery_hook = mock_bq_hook
        op.bigquery_hook.project_id = TEST_PROJECT
        op.mysql_hook = mock_mysql_hook
        context = mock.MagicMock()
        op.execute(context=context)

        result = op.get_openlineage_facets_on_complete(None)
        assert len(result.inputs) == 1
        assert len(result.outputs) == 1

        input_ds = result.inputs[0]
        assert input_ds.namespace == "bigquery"
        assert input_ds.name == f"{TEST_PROJECT}.{TEST_DATASET}.{TEST_TABLE_ID}"
        assert "schema" in input_ds.facets
        schema_fields = [f.name for f in input_ds.facets["schema"].fields]
        assert set(schema_fields) == {"id", "name", "value"}

        output_ds = result.outputs[0]
        assert output_ds.namespace == "mysql://localhost:3306"
        assert output_ds.name == "mydb.destination"
        assert "columnLineage" in output_ds.facets
        col_lineage = output_ds.facets["columnLineage"]
        assert set(col_lineage.fields.keys()) == {"id", "name", "value"}

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_sql.BigQueryHook")
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_mysql.MySqlHook")
    def test_get_openlineage_facets_on_complete_selected_fields(self, mock_mysql_hook, mock_bq_hook):
        mock_bq_client = MagicMock()
        mock_bq_client.get_table.return_value = _make_bq_table(["id", "name", "value"])
        mock_bq_hook.get_client.return_value = mock_bq_client
        mock_bq_hook.return_value = mock_bq_hook

        db_info = MagicMock(scheme="mysql", authority="localhost:3306", database="mydb")
        mock_mysql_hook.get_openlineage_database_info.return_value = db_info
        mock_mysql_hook.return_value = mock_mysql_hook

        op = BigQueryToMySqlOperator(
            task_id=TASK_ID,
            dataset_table=f"{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name="destination",
            selected_fields=["id", "name"],
            database="mydb",
        )
        op.bigquery_hook = mock_bq_hook
        op.bigquery_hook.project_id = TEST_PROJECT
        op.mysql_hook = mock_mysql_hook
        context = mock.MagicMock()
        op.execute(context=context)

        result = op.get_openlineage_facets_on_complete(None)
        assert len(result.inputs) == 1
        assert len(result.outputs) == 1

        input_ds = result.inputs[0]
        assert input_ds.namespace == "bigquery"
        assert input_ds.name == f"{TEST_PROJECT}.{TEST_DATASET}.{TEST_TABLE_ID}"
        assert "schema" in input_ds.facets
        schema_fields = [f.name for f in input_ds.facets["schema"].fields]
        assert set(schema_fields) == {"id", "name"}

        output_ds = result.outputs[0]
        assert output_ds.namespace == "mysql://localhost:3306"
        assert output_ds.name == "mydb.destination"
        assert "columnLineage" in output_ds.facets
        col_lineage = output_ds.facets["columnLineage"]
        assert set(col_lineage.fields.keys()) == {"id", "name"}



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import os
from unittest import mock
from unittest.mock import patch

import pytest
from requests import HTTPError
from tenacity import RetryError

from airflow.providers.google.cloud.hooks.dataprep import GoogleDataprepHook

try:
    import importlib.util

    if not importlib.util.find_spec("airflow.sdk.bases.hook"):
        raise ImportError

    BASEHOOK_PATCH_PATH = "airflow.sdk.bases.hook.BaseHook"
except ImportError:
    BASEHOOK_PATCH_PATH = "airflow.hooks.base.BaseHook"
JOB_ID = 1234567
RECIPE_ID = 1234567
TOKEN = "1111"
EXTRA = {"token": TOKEN}
EMBED = ""
INCLUDE_DELETED = False
DATA = {"wrangledDataset": {"id": RECIPE_ID}}
URL_BASE = "https://api.clouddataprep.com"
URL_JOB_GROUPS = URL_BASE + "/v4/jobGroups"
URL_IMPORTED_DATASETS = URL_BASE + "/v4/importedDatasets"
URL_WRANGLED_DATASETS = URL_BASE + "/v4/wrangledDatasets"
URL_OUTPUT_OBJECTS = URL_BASE + "/v4/outputObjects"
URL_WRITE_SETTINGS = URL_BASE + "/v4/writeSettings"

# needed to ignore MyPy badly detecting HTTPError as requiring response parameter
# HTTPError will fall-back to None when parameter is not present.
# mypy: disable-error-code="call-arg"


class TestGoogleDataprepHook:
    def setup_method(self):
        with mock.patch(f"{BASEHOOK_PATCH_PATH}.get_connection") as conn:
            conn.return_value.extra_dejson = EXTRA
            self.hook = GoogleDataprepHook(dataprep_conn_id="dataprep_default")
        self._imported_dataset_id = 12345
        self._create_imported_dataset_body_request = {
            "uri": "gs://test/uri",
            "name": "test_name",
        }
        self._create_wrangled_dataset_body_request = {
            "importedDataset": {"id": "test_dataset_id"},
            "flow": {"id": "test_flow_id"},
            "name": "test_dataset_name",
        }
        self._create_output_object_body_request = {
            "execution": "dataflow",
            "profiler": False,
            "flowNodeId": "test_flow_node_id",
        }
        self._create_write_settings_body_request = {
            "path": "gs://test/path",
            "action": "create",
            "format": "csv",
            "outputObjectId": "test_output_object_id",
        }
        self._expected_create_imported_dataset_hook_data = json.dumps(
            self._create_imported_dataset_body_request
        )
        self._expected_create_wrangled_dataset_hook_data = json.dumps(
            self._create_wrangled_dataset_body_request
        )
        self._expected_create_output_object_hook_data = json.dumps(self._create_output_object_body_request)
        self._expected_create_write_settings_hook_data = json.dumps(self._create_write_settings_body_request)

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.get")
    def test_get_jobs_for_job_group_should_be_called_once_with_params(self, mock_get_request):
        self.hook.get_jobs_for_job_group(JOB_ID)
        mock_get_request.assert_called_once_with(
            f"{URL_JOB_GROUPS}/{JOB_ID}/jobs",
            headers={"Content-Type": "application/json", "Authorization": f"Bearer {TOKEN}"},
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_get_jobs_for_job_group_should_pass_after_retry(self, mock_get_request):
        self.hook.get_jobs_for_job_group(JOB_ID)
        assert mock_get_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_get_jobs_for_job_group_should_not_retry_after_success(self, mock_get_request):
        self.hook.get_jobs_for_job_group.retry.sleep = mock.Mock()
        self.hook.get_jobs_for_job_group(JOB_ID)
        assert mock_get_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), mock.MagicMock()],
    )
    def test_get_jobs_for_job_group_should_retry_after_four_errors(self, mock_get_request):
        self.hook.get_jobs_for_job_group.retry.sleep = mock.Mock()
        self.hook.get_jobs_for_job_group(JOB_ID)
        assert mock_get_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_get_jobs_for_job_group_raise_error_after_five_calls(self, mock_get_request):
        self.hook.get_jobs_for_job_group.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.get_jobs_for_job_group(JOB_ID)
        assert "HTTPError" in str(ctx.value)
        assert mock_get_request.call_count == 5

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.get")
    def test_get_job_group_should_be_called_once_with_params(self, mock_get_request):
        self.hook.get_job_group(JOB_ID, EMBED, INCLUDE_DELETED)
        mock_get_request.assert_called_once_with(
            f"{URL_JOB_GROUPS}/{JOB_ID}",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
            params={"embed": "", "includeDeleted": False},
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_get_job_group_should_pass_after_retry(self, mock_get_request):
        self.hook.get_job_group(JOB_ID, EMBED, INCLUDE_DELETED)
        assert mock_get_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_get_job_group_should_not_retry_after_success(self, mock_get_request):
        self.hook.get_job_group.retry.sleep = mock.Mock()
        self.hook.get_job_group(JOB_ID, EMBED, INCLUDE_DELETED)
        assert mock_get_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_get_job_group_should_retry_after_four_errors(self, mock_get_request):
        self.hook.get_job_group.retry.sleep = mock.Mock()
        self.hook.get_job_group(JOB_ID, EMBED, INCLUDE_DELETED)
        assert mock_get_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_get_job_group_raise_error_after_five_calls(self, mock_get_request):
        self.hook.get_job_group.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.get_job_group(JOB_ID, EMBED, INCLUDE_DELETED)
        assert "HTTPError" in str(ctx.value)
        assert mock_get_request.call_count == 5

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.post")
    def test_run_job_group_should_be_called_once_with_params(self, mock_get_request):
        self.hook.run_job_group(body_request=DATA)
        mock_get_request.assert_called_once_with(
            f"{URL_JOB_GROUPS}",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
            data=json.dumps(DATA),
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_run_job_group_should_pass_after_retry(self, mock_get_request):
        self.hook.run_job_group(body_request=DATA)
        assert mock_get_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_run_job_group_should_not_retry_after_success(self, mock_get_request):
        self.hook.run_job_group.retry.sleep = mock.Mock()
        self.hook.run_job_group(body_request=DATA)
        assert mock_get_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_run_job_group_should_retry_after_four_errors(self, mock_get_request):
        self.hook.run_job_group.retry.sleep = mock.Mock()
        self.hook.run_job_group(body_request=DATA)
        assert mock_get_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_run_job_group_raise_error_after_five_calls(self, mock_get_request):
        self.hook.run_job_group.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.run_job_group(body_request=DATA)
        assert "HTTPError" in str(ctx.value)
        assert mock_get_request.call_count == 5

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.get")
    def test_get_job_group_status_should_be_called_once_with_params(self, mock_get_request):
        self.hook.get_job_group_status(job_group_id=JOB_ID)
        mock_get_request.assert_called_once_with(
            f"{URL_JOB_GROUPS}/{JOB_ID}/status",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_get_job_group_status_should_pass_after_retry(self, mock_get_request):
        self.hook.get_job_group_status(job_group_id=JOB_ID)
        assert mock_get_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_get_job_group_status_retry_after_success(self, mock_get_request):
        self.hook.run_job_group.retry.sleep = mock.Mock()
        self.hook.get_job_group_status(job_group_id=JOB_ID)
        assert mock_get_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_get_job_group_status_four_errors(self, mock_get_request):
        self.hook.run_job_group.retry.sleep = mock.Mock()
        self.hook.get_job_group_status(job_group_id=JOB_ID)
        assert mock_get_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.get",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_get_job_group_status_five_calls(self, mock_get_request):
        self.hook.get_job_group_status.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.get_job_group_status(job_group_id=JOB_ID)
        assert "HTTPError" in str(ctx.value)
        assert mock_get_request.call_count == 5

    @pytest.mark.parametrize(
        "uri",
        [
            pytest.param("a://?extra__dataprep__token=abc&extra__dataprep__base_url=abc", id="prefix"),
            pytest.param("a://?token=abc&base_url=abc", id="no-prefix"),
        ],
    )
    def test_conn_extra_backcompat_prefix(self, uri):
        with patch.dict(os.environ, {"AIRFLOW_CONN_MY_CONN": uri}):
            hook = GoogleDataprepHook("my_conn")
            assert hook._token == "abc"
            assert hook._base_url == "abc"

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.post")
    def test_create_imported_dataset_should_be_called_once_with_params(self, mock_post_request):
        self.hook.create_imported_dataset(body_request=self._create_imported_dataset_body_request)
        mock_post_request.assert_called_once_with(
            URL_IMPORTED_DATASETS,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
            data=self._expected_create_imported_dataset_hook_data,
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_create_imported_dataset_should_pass_after_retry(self, mock_post_request):
        self.hook.create_imported_dataset(body_request=self._create_imported_dataset_body_request)
        assert mock_post_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_create_imported_dataset_retry_after_success(self, mock_post_request):
        self.hook.create_imported_dataset.retry.sleep = mock.Mock()
        self.hook.create_imported_dataset(body_request=self._create_imported_dataset_body_request)
        assert mock_post_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_create_imported_dataset_four_errors(self, mock_post_request):
        self.hook.create_imported_dataset.retry.sleep = mock.Mock()
        self.hook.create_imported_dataset(body_request=self._create_imported_dataset_body_request)
        assert mock_post_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_create_imported_dataset_five_calls(self, mock_post_request):
        self.hook.create_imported_dataset.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.create_imported_dataset(body_request=self._create_imported_dataset_body_request)
        assert "HTTPError" in str(ctx.value)
        assert mock_post_request.call_count == 5

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.post")
    def test_create_wrangled_dataset_should_be_called_once_with_params(self, mock_post_request):
        self.hook.create_wrangled_dataset(body_request=self._create_wrangled_dataset_body_request)
        mock_post_request.assert_called_once_with(
            URL_WRANGLED_DATASETS,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
            data=self._expected_create_wrangled_dataset_hook_data,
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_create_wrangled_dataset_should_pass_after_retry(self, mock_post_request):
        self.hook.create_wrangled_dataset(body_request=self._create_wrangled_dataset_body_request)
        assert mock_post_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_create_wrangled_dataset_retry_after_success(self, mock_post_request):
        self.hook.create_wrangled_dataset.retry.sleep = mock.Mock()
        self.hook.create_wrangled_dataset(body_request=self._create_wrangled_dataset_body_request)
        assert mock_post_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_create_wrangled_dataset_four_errors(self, mock_post_request):
        self.hook.create_wrangled_dataset.retry.sleep = mock.Mock()
        self.hook.create_wrangled_dataset(body_request=self._create_wrangled_dataset_body_request)
        assert mock_post_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_create_wrangled_dataset_five_calls(self, mock_post_request):
        self.hook.create_wrangled_dataset.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.create_wrangled_dataset(body_request=self._create_wrangled_dataset_body_request)
        assert "HTTPError" in str(ctx.value)
        assert mock_post_request.call_count == 5

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.post")
    def test_create_output_object_should_be_called_once_with_params(self, mock_post_request):
        self.hook.create_output_object(body_request=self._create_output_object_body_request)
        mock_post_request.assert_called_once_with(
            URL_OUTPUT_OBJECTS,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
            data=self._expected_create_output_object_hook_data,
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_create_output_objects_should_pass_after_retry(self, mock_post_request):
        self.hook.create_output_object(body_request=self._create_output_object_body_request)
        assert mock_post_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_create_output_objects_retry_after_success(self, mock_post_request):
        self.hook.create_output_object.retry.sleep = mock.Mock()
        self.hook.create_output_object(body_request=self._create_output_object_body_request)
        assert mock_post_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_create_output_objects_four_errors(self, mock_post_request):
        self.hook.create_output_object.retry.sleep = mock.Mock()
        self.hook.create_output_object(body_request=self._create_output_object_body_request)
        assert mock_post_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_create_output_objects_five_calls(self, mock_post_request):
        self.hook.create_output_object.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.create_output_object(body_request=self._create_output_object_body_request)
        assert "HTTPError" in str(ctx.value)
        assert mock_post_request.call_count == 5

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.post")
    def test_create_write_settings_should_be_called_once_with_params(self, mock_post_request):
        self.hook.create_write_settings(body_request=self._create_write_settings_body_request)
        mock_post_request.assert_called_once_with(
            URL_WRITE_SETTINGS,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
            data=self._expected_create_write_settings_hook_data,
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_create_write_settings_should_pass_after_retry(self, mock_post_request):
        self.hook.create_write_settings(body_request=self._create_write_settings_body_request)
        assert mock_post_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_create_write_settings_retry_after_success(self, mock_post_request):
        self.hook.create_write_settings.retry.sleep = mock.Mock()
        self.hook.create_write_settings(body_request=self._create_write_settings_body_request)
        assert mock_post_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_create_write_settings_four_errors(self, mock_post_request):
        self.hook.create_write_settings.retry.sleep = mock.Mock()
        self.hook.create_write_settings(body_request=self._create_write_settings_body_request)
        assert mock_post_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_create_write_settings_five_calls(self, mock_post_request):
        self.hook.create_write_settings.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.create_write_settings(body_request=self._create_write_settings_body_request)
        assert "HTTPError" in str(ctx.value)
        assert mock_post_request.call_count == 5

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.delete")
    def test_delete_imported_dataset_should_be_called_once_with_params(self, mock_delete_request):
        self.hook.delete_imported_dataset(dataset_id=self._imported_dataset_id)
        mock_delete_request.assert_called_once_with(
            f"{URL_IMPORTED_DATASETS}/{self._imported_dataset_id}",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.delete",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_delete_imported_dataset_should_pass_after_retry(self, mock_delete_request):
        self.hook.delete_imported_dataset(dataset_id=self._imported_dataset_id)
        assert mock_delete_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.delete",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_delete_imported_dataset_retry_after_success(self, mock_delete_request):
        self.hook.delete_imported_dataset.retry.sleep = mock.Mock()
        self.hook.delete_imported_dataset(dataset_id=self._imported_dataset_id)
        assert mock_delete_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.delete",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_delete_imported_dataset_four_errors(self, mock_delete_request):
        self.hook.delete_imported_dataset.retry.sleep = mock.Mock()
        self.hook.delete_imported_dataset(dataset_id=self._imported_dataset_id)
        assert mock_delete_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.delete",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_delete_imported_dataset_five_calls(self, mock_delete_request):
        self.hook.delete_imported_dataset.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.delete_imported_dataset(dataset_id=self._imported_dataset_id)
        assert "HTTPError" in str(ctx.value)
        assert mock_delete_request.call_count == 5


class TestGoogleDataprepFlowPathHooks:
    _url = "https://api.clouddataprep.com/v4/flows"

    def setup_method(self):
        self._flow_id = 1234567
        self._create_flow_body_request = {
            "name": "test_name",
            "description": "Test description",
        }
        self._expected_copy_flow_hook_data = json.dumps(
            {
                "name": "",
                "description": "",
                "copyDatasources": False,
            }
        )
        self._expected_run_flow_hook_data = json.dumps({})
        self._expected_create_flow_hook_data = json.dumps(
            {
                "name": "test_name",
                "description": "Test description",
            }
        )
        with mock.patch(f"{BASEHOOK_PATCH_PATH}.get_connection") as conn:
            conn.return_value.extra_dejson = EXTRA
            self.hook = GoogleDataprepHook(dataprep_conn_id="dataprep_default")

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.post")
    def test_create_flow_should_be_called_once_with_params(self, mock_post_request):
        self.hook.create_flow(body_request=self._create_flow_body_request)
        mock_post_request.assert_called_once_with(
            self._url,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
            data=self._expected_create_flow_hook_data,
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_create_flow_should_pass_after_retry(self, mock_post_request):
        self.hook.create_flow(body_request=self._create_flow_body_request)
        assert mock_post_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_create_flow_should_not_retry_after_success(self, mock_post_request):
        self.hook.create_flow.retry.sleep = mock.Mock()
        self.hook.create_flow(body_request=self._create_flow_body_request)
        assert mock_post_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_create_flow_should_retry_after_four_errors(self, mock_post_request):
        self.hook.create_flow.retry.sleep = mock.Mock()
        self.hook.create_flow(body_request=self._create_flow_body_request)
        assert mock_post_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_create_flow_raise_error_after_five_calls(self, mock_post_request):
        self.hook.create_flow.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.create_flow(body_request=self._create_flow_body_request)
        assert "HTTPError" in str(ctx.value)
        assert mock_post_request.call_count == 5

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.post")
    def test_copy_flow_should_be_called_once_with_params(self, mock_get_request):
        self.hook.copy_flow(
            flow_id=self._flow_id,
        )
        mock_get_request.assert_called_once_with(
            f"{self._url}/{self._flow_id}/copy",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
            data=self._expected_copy_flow_hook_data,
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_copy_flow_should_pass_after_retry(self, mock_get_request):
        self.hook.copy_flow(flow_id=self._flow_id)
        assert mock_get_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_copy_flow_should_not_retry_after_success(self, mock_get_request):
        self.hook.copy_flow.retry.sleep = mock.Mock()
        self.hook.copy_flow(flow_id=self._flow_id)
        assert mock_get_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_copy_flow_should_retry_after_four_errors(self, mock_get_request):
        self.hook.copy_flow.retry.sleep = mock.Mock()
        self.hook.copy_flow(flow_id=self._flow_id)
        assert mock_get_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_copy_flow_raise_error_after_five_calls(self, mock_get_request):
        self.hook.copy_flow.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.copy_flow(flow_id=self._flow_id)
        assert "HTTPError" in str(ctx.value)
        assert mock_get_request.call_count == 5

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.delete")
    def test_delete_flow_should_be_called_once_with_params(self, mock_get_request):
        self.hook.delete_flow(
            flow_id=self._flow_id,
        )
        mock_get_request.assert_called_once_with(
            f"{self._url}/{self._flow_id}",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.delete",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_delete_flow_should_pass_after_retry(self, mock_get_request):
        self.hook.delete_flow(flow_id=self._flow_id)
        assert mock_get_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.delete",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_delete_flow_should_not_retry_after_success(self, mock_get_request):
        self.hook.delete_flow.retry.sleep = mock.Mock()
        self.hook.delete_flow(flow_id=self._flow_id)
        assert mock_get_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.delete",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_delete_flow_should_retry_after_four_errors(self, mock_get_request):
        self.hook.delete_flow.retry.sleep = mock.Mock()
        self.hook.delete_flow(flow_id=self._flow_id)
        assert mock_get_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.delete",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_delete_flow_raise_error_after_five_calls(self, mock_get_request):
        self.hook.delete_flow.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.delete_flow(flow_id=self._flow_id)
        assert "HTTPError" in str(ctx.value)
        assert mock_get_request.call_count == 5

    @patch("airflow.providers.google.cloud.hooks.dataprep.requests.post")
    def test_run_flow_should_be_called_once_with_params(self, mock_get_request):
        self.hook.run_flow(
            flow_id=self._flow_id,
            body_request={},
        )
        mock_get_request.assert_called_once_with(
            f"{self._url}/{self._flow_id}/run",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {TOKEN}",
            },
            data=self._expected_run_flow_hook_data,
        )

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), mock.MagicMock()],
    )
    def test_run_flow_should_pass_after_retry(self, mock_get_request):
        self.hook.run_flow(
            flow_id=self._flow_id,
            body_request={},
        )
        assert mock_get_request.call_count == 2

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[mock.MagicMock(), HTTPError()],
    )
    def test_run_flow_should_not_retry_after_success(self, mock_get_request):
        self.hook.run_flow.retry.sleep = mock.Mock()
        self.hook.run_flow(
            flow_id=self._flow_id,
            body_request={},
        )
        assert mock_get_request.call_count == 1

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[
            HTTPError(),
            HTTPError(),
            HTTPError(),
            HTTPError(),
            mock.MagicMock(),
        ],
    )
    def test_run_flow_should_retry_after_four_errors(self, mock_get_request):
        self.hook.run_flow.retry.sleep = mock.Mock()
        self.hook.run_flow(
            flow_id=self._flow_id,
            body_request={},
        )
        assert mock_get_request.call_count == 5

    @patch(
        "airflow.providers.google.cloud.hooks.dataprep.requests.post",
        side_effect=[HTTPError(), HTTPError(), HTTPError(), HTTPError(), HTTPError()],
    )
    def test_run_flow_raise_error_after_five_calls(self, mock_get_request):
        self.hook.run_flow.retry.sleep = mock.Mock()
        with pytest.raises(RetryError) as ctx:
            self.hook.run_flow(
                flow_id=self._flow_id,
                body_request={},
            )
        assert "HTTPError" in str(ctx.value)
        assert mock_get_request.call_count == 5



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.marketing_platform.operators.search_ads import (
    GoogleSearchAdsGetCustomColumnOperator,
    GoogleSearchAdsGetFieldOperator,
    GoogleSearchAdsListCustomColumnsOperator,
    GoogleSearchAdsSearchFieldsOperator,
    GoogleSearchAdsSearchOperator,
)

GCP_CONN_ID = "google_search_ads_default"
API_VERSION = "v0"
CUSTOMER_ID = "customer_id"


class TestGoogleSearchAdsSearchOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.search_ads.GoogleSearchAdsReportingHook"
    )
    @mock.patch("airflow.providers.google.marketing_platform.operators.search_ads.BaseOperator")
    def test_execute(self, mock_base_op, hook_mock):
        query = "SELECT * FROM campaigns WHERE segments.date DURING LAST_30_DAYS"
        hook_mock.return_value.search.return_value = {"results": []}
        op = GoogleSearchAdsSearchOperator(
            customer_id=CUSTOMER_ID,
            query=query,
            api_version=API_VERSION,
            task_id="test_task",
        )
        op.execute(context=None)
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version="v0",
        )
        hook_mock.return_value.search.assert_called_once_with(
            customer_id=CUSTOMER_ID,
            query=query,
            page_size=10000,
            page_token=None,
            return_total_results_count=False,
            summary_row_setting=None,
            validate_only=False,
        )


class TestGoogleSearchAdsGetFieldOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.search_ads.GoogleSearchAdsReportingHook"
    )
    @mock.patch("airflow.providers.google.marketing_platform.operators.search_ads.BaseOperator")
    def test_execute(self, mock_base_op, hook_mock):
        field_name = "the_field"
        hook_mock.return_value.get_field.return_value = {
            "name": field_name,
            "resourceName": f"searchAds360Fields/{field_name}",
        }
        op = GoogleSearchAdsGetFieldOperator(
            field_name=field_name,
            api_version=API_VERSION,
            task_id="test_task",
        )
        op.execute(context=None)
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version="v0",
        )
        hook_mock.return_value.get_field.assert_called_once_with(
            field_name=field_name,
        )


class TestGoogleSearchAdsSearchFieldsOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.search_ads.GoogleSearchAdsReportingHook"
    )
    @mock.patch("airflow.providers.google.marketing_platform.operators.search_ads.BaseOperator")
    def test_execute(self, mock_base_op, hook_mock):
        field_name = "the_field"
        query = (
            "SELECT "
            "  name, category, selectable, filterable, sortable, selectable_with, data_type, "
            "  is_repeated "
            "WHERE "
            "  name LIKE 'ad_group.%'"
        )
        hook_mock.return_value.search_fields.return_value = {
            "results": [{"name": field_name, "resource_name": f"searchAds360Fields/{field_name}"}]
        }
        op = GoogleSearchAdsSearchFieldsOperator(
            query=query,
            api_version=API_VERSION,
            task_id="test_task",
        )
        op.execute(context=None)
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version="v0",
        )
        hook_mock.return_value.search_fields.assert_called_once_with(
            query=query,
            page_token=None,
            page_size=10000,
        )


class TestGoogleSearchAdsGetCustomColumnOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.search_ads.GoogleSearchAdsReportingHook"
    )
    @mock.patch("airflow.providers.google.marketing_platform.operators.search_ads.BaseOperator")
    def test_execute(self, mock_base_op, hook_mock):
        custom_column_id = "custom_column_id"
        hook_mock.return_value.get_custom_column.return_value = {"id": custom_column_id}
        op = GoogleSearchAdsGetCustomColumnOperator(
            customer_id=CUSTOMER_ID,
            custom_column_id=custom_column_id,
            api_version=API_VERSION,
            task_id="test_task",
        )
        op.execute(context=None)
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version="v0",
        )
        hook_mock.return_value.get_custom_column.assert_called_once_with(
            customer_id=CUSTOMER_ID,
            custom_column_id=custom_column_id,
        )


class TestGoogleSearchAdsListCustomColumnsOperator:
    @mock.patch(
        "airflow.providers.google.marketing_platform.operators.search_ads.GoogleSearchAdsReportingHook"
    )
    @mock.patch("airflow.providers.google.marketing_platform.operators.search_ads.BaseOperator")
    def test_execute(self, mock_base_op, hook_mock):
        customs_columns = [
            {"id": "custom_column_id_1"},
            {"id": "custom_column_id_2"},
            {"id": "custom_column_id_3"},
        ]
        hook_mock.return_value.list_custom_columns.return_value = {"customColumns": customs_columns}
        op = GoogleSearchAdsListCustomColumnsOperator(
            customer_id=CUSTOMER_ID,
            api_version=API_VERSION,
            task_id="test_task",
        )
        op.execute(context=None)
        hook_mock.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            api_version="v0",
        )
        hook_mock.return_value.list_custom_columns.assert_called_once_with(
            customer_id=CUSTOMER_ID,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow import DAG
from airflow.providers.common.compat.sdk import AirflowException, TaskDeferred
from airflow.providers.google.cloud.hooks.datafusion import SUCCESS_STATES, PipelineStates
from airflow.providers.google.cloud.openlineage.facets import DataFusionRunFacet
from airflow.providers.google.cloud.operators.datafusion import (
    CloudDataFusionCreateInstanceOperator,
    CloudDataFusionCreatePipelineOperator,
    CloudDataFusionDeleteInstanceOperator,
    CloudDataFusionDeletePipelineOperator,
    CloudDataFusionGetInstanceOperator,
    CloudDataFusionListPipelinesOperator,
    CloudDataFusionRestartInstanceOperator,
    CloudDataFusionStartPipelineOperator,
    CloudDataFusionStopPipelineOperator,
    CloudDataFusionUpdateInstanceOperator,
)
from airflow.providers.google.cloud.triggers.datafusion import DataFusionStartPipelineTrigger
from airflow.providers.google.cloud.utils.datafusion import DataFusionPipelineType

HOOK_STR = "airflow.providers.google.cloud.operators.datafusion.DataFusionHook"
RESOURCE_PATH_TO_DICT_STR = "airflow.providers.google.cloud.operators.datafusion.resource_path_to_dict"

TASK_ID = "test_task"
LOCATION = "test-location"
INSTANCE_NAME = "airflow-test-instance"
INSTANCE = {"type": "BASIC", "displayName": INSTANCE_NAME}
PROJECT_ID = "test_project_id"
PIPELINE_NAME = "shrubberyPipeline"
PIPELINE = {"test": "pipeline"}
PIPELINE_ID = "test_pipeline_id"
INSTANCE_URL = "http://datafusion.instance.com"
NAMESPACE = "TEST_NAMESPACE"
RUNTIME_ARGS = {"arg1": "a", "arg2": "b"}


class TestCloudDataFusionUpdateInstanceOperator:
    @mock.patch(RESOURCE_PATH_TO_DICT_STR)
    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_should_execute_successfully(self, mock_hook, mock_resource_to_dict):
        update_maks = "instance.name"
        mock_resource_to_dict.return_value = {"projects": PROJECT_ID}
        op = CloudDataFusionUpdateInstanceOperator(
            task_id="test_tasks",
            instance_name=INSTANCE_NAME,
            instance=INSTANCE,
            update_mask=update_maks,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        op.execute(context=mock.MagicMock())

        mock_hook.return_value.patch_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME,
            instance=INSTANCE,
            update_mask=update_maks,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        assert mock_hook.return_value.wait_for_operation.call_count == 1


class TestCloudDataFusionRestartInstanceOperator:
    @mock.patch(RESOURCE_PATH_TO_DICT_STR)
    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_should_execute_successfully(self, mock_hook, mock_resource_path_to_dict):
        mock_resource_path_to_dict.return_value = {"projects": PROJECT_ID}
        op = CloudDataFusionRestartInstanceOperator(
            task_id="test_tasks",
            instance_name=INSTANCE_NAME,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        op.execute(context=mock.MagicMock())

        mock_hook.return_value.restart_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )
        assert mock_hook.return_value.wait_for_operation.call_count == 1


class TestCloudDataFusionCreateInstanceOperator:
    @mock.patch(RESOURCE_PATH_TO_DICT_STR)
    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_should_execute_successfully(self, mock_hook, mock_resource_path_to_dict):
        mock_resource_path_to_dict.return_value = {"projects": PROJECT_ID}
        op = CloudDataFusionCreateInstanceOperator(
            task_id="test_tasks",
            instance_name=INSTANCE_NAME,
            instance=INSTANCE,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        op.execute(context=mock.MagicMock())

        mock_hook.return_value.create_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME,
            instance=INSTANCE,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        assert mock_hook.return_value.wait_for_operation.call_count == 1


class TestCloudDataFusionDeleteInstanceOperator:
    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_should_execute_successfully(self, mock_hook):
        op = CloudDataFusionDeleteInstanceOperator(
            task_id="test_tasks",
            instance_name=INSTANCE_NAME,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        op.execute(context=mock.MagicMock())

        mock_hook.return_value.delete_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )
        assert mock_hook.return_value.wait_for_operation.call_count == 1


class TestCloudDataFusionGetInstanceOperator:
    @mock.patch(RESOURCE_PATH_TO_DICT_STR)
    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_should_execute_successfully(self, mock_hook, mock_resource_path_to_dict):
        mock_resource_path_to_dict.return_value = {"projects": PROJECT_ID}
        op = CloudDataFusionGetInstanceOperator(
            task_id="test_tasks",
            instance_name=INSTANCE_NAME,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        op.execute(context=mock.MagicMock())

        mock_hook.return_value.get_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )


class TestCloudDataFusionCreatePipelineOperator:
    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_should_execute_successfully(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {
            "apiEndpoint": INSTANCE_URL,
            "serviceEndpoint": INSTANCE_URL,
        }
        op = CloudDataFusionCreatePipelineOperator(
            task_id="test_tasks",
            pipeline_name=PIPELINE_NAME,
            pipeline=PIPELINE,
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.return_value.get_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )

        mock_hook.return_value.create_pipeline.assert_called_once_with(
            instance_url=INSTANCE_URL,
            pipeline_name=PIPELINE_NAME,
            pipeline=PIPELINE,
            namespace=NAMESPACE,
        )


class TestCloudDataFusionDeletePipelineOperator:
    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_should_execute_successfully(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {
            "apiEndpoint": INSTANCE_URL,
            "serviceEndpoint": INSTANCE_URL,
        }
        op = CloudDataFusionDeletePipelineOperator(
            task_id="test_tasks",
            pipeline_name=PIPELINE_NAME,
            version_id="1.12",
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.return_value.get_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )

        mock_hook.return_value.delete_pipeline.assert_called_once_with(
            instance_url=INSTANCE_URL,
            pipeline_name=PIPELINE_NAME,
            namespace=NAMESPACE,
            version_id="1.12",
        )


class TestCloudDataFusionStartPipelineOperator:
    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_should_execute_successfully(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {
            "apiEndpoint": INSTANCE_URL,
            "serviceEndpoint": INSTANCE_URL,
        }
        mock_hook.return_value.start_pipeline.return_value = PIPELINE_ID

        op = CloudDataFusionStartPipelineOperator(
            task_id=TASK_ID,
            pipeline_name=PIPELINE_NAME,
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
            runtime_args=RUNTIME_ARGS,
        )
        op.dag = mock.MagicMock(spec=DAG, task_dict={}, dag_id="test")

        op.execute(context=mock.MagicMock())
        mock_hook.return_value.get_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )

        mock_hook.return_value.start_pipeline.assert_called_once_with(
            instance_url=INSTANCE_URL,
            pipeline_name=PIPELINE_NAME,
            namespace=NAMESPACE,
            runtime_args=RUNTIME_ARGS,
            pipeline_type=DataFusionPipelineType.BATCH,
        )

        mock_hook.return_value.wait_for_pipeline_state.assert_called_once_with(
            success_states=[*SUCCESS_STATES, PipelineStates.RUNNING],
            pipeline_id=PIPELINE_ID,
            pipeline_name=PIPELINE_NAME,
            pipeline_type=DataFusionPipelineType.BATCH,
            namespace=NAMESPACE,
            instance_url=INSTANCE_URL,
            timeout=300,
        )

    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_asynch_param_should_execute_successfully(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {
            "apiEndpoint": INSTANCE_URL,
            "serviceEndpoint": INSTANCE_URL,
        }
        mock_hook.return_value.start_pipeline.return_value = PIPELINE_ID

        op = CloudDataFusionStartPipelineOperator(
            task_id=TASK_ID,
            pipeline_name=PIPELINE_NAME,
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
            runtime_args=RUNTIME_ARGS,
            asynchronous=True,
        )
        op.dag = mock.MagicMock(spec=DAG, task_dict={}, dag_id="test")
        op.execute(context=mock.MagicMock())

        mock_hook.return_value.get_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )
        mock_hook.return_value.start_pipeline.assert_called_once_with(
            instance_url=INSTANCE_URL,
            pipeline_name=PIPELINE_NAME,
            namespace=NAMESPACE,
            runtime_args=RUNTIME_ARGS,
            pipeline_type=DataFusionPipelineType.BATCH,
        )
        mock_hook.return_value.wait_for_pipeline_state.assert_not_called()


class TestCloudDataFusionStartPipelineOperatorAsync:
    @mock.patch(HOOK_STR)
    def test_asynch_execute_should_execute_successfully(self, mock_hook):
        """
        Asserts that a task is deferred and a DataFusionStartPipelineTrigger will be fired
        when the CloudDataFusionStartPipelineOperator is executed in deferrable mode when deferrable=True.
        """

        op = CloudDataFusionStartPipelineOperator(
            task_id=TASK_ID,
            pipeline_name=PIPELINE_NAME,
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
            runtime_args=RUNTIME_ARGS,
            deferrable=True,
        )
        op.dag = mock.MagicMock(spec=DAG, task_dict={}, dag_id="test")
        with pytest.raises(TaskDeferred) as exc:
            op.execute(context=mock.MagicMock())

        assert isinstance(exc.value.trigger, DataFusionStartPipelineTrigger), (
            "Trigger is not a DataFusionStartPipelineTrigger"
        )

    def test_asynch_execute_should_should_throw_exception(self):
        """Tests that an AirflowException is raised in case of error event"""

        op = CloudDataFusionStartPipelineOperator(
            task_id=TASK_ID,
            pipeline_name=PIPELINE_NAME,
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
            runtime_args=RUNTIME_ARGS,
            deferrable=True,
        )
        with pytest.raises(AirflowException):
            op.execute_complete(
                context=mock.MagicMock(), event={"status": "error", "message": "test failure message"}
            )

    def test_asynch_execute_logging_should_execute_successfully(self):
        """Asserts that logging occurs as expected"""

        op = CloudDataFusionStartPipelineOperator(
            task_id=TASK_ID,
            pipeline_name=PIPELINE_NAME,
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
            runtime_args=RUNTIME_ARGS,
            deferrable=True,
        )
        with mock.patch.object(op.log, "info") as mock_log_info:
            op.execute_complete(
                context=mock.MagicMock(),
                event={"status": "success", "message": "Pipeline completed", "pipeline_id": PIPELINE_ID},
            )
        mock_log_info.assert_called_with("%s completed with response %s ", TASK_ID, "Pipeline completed")

    @mock.patch(HOOK_STR)
    def test_asynch_execute_check_hook_call_should_execute_successfully(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {
            "apiEndpoint": INSTANCE_URL,
            "serviceEndpoint": INSTANCE_URL,
        }
        mock_hook.return_value.start_pipeline.return_value = PIPELINE_ID
        op = CloudDataFusionStartPipelineOperator(
            task_id=TASK_ID,
            pipeline_name=PIPELINE_NAME,
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
            runtime_args=RUNTIME_ARGS,
            deferrable=True,
        )
        with pytest.raises(TaskDeferred):
            op.execute(context=mock.MagicMock())

        mock_hook.return_value.get_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )
        mock_hook.return_value.start_pipeline.assert_called_once_with(
            instance_url=INSTANCE_URL,
            pipeline_name=PIPELINE_NAME,
            namespace=NAMESPACE,
            runtime_args=RUNTIME_ARGS,
            pipeline_type=DataFusionPipelineType.BATCH,
        )

    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_asynch_param_should_execute_successfully(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {
            "apiEndpoint": INSTANCE_URL,
            "serviceEndpoint": INSTANCE_URL,
        }
        mock_hook.return_value.start_pipeline.return_value = PIPELINE_ID
        op = CloudDataFusionStartPipelineOperator(
            task_id=TASK_ID,
            pipeline_name=PIPELINE_NAME,
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
            runtime_args=RUNTIME_ARGS,
            asynchronous=True,
            deferrable=True,
        )
        op.dag = mock.MagicMock(spec=DAG, task_dict={}, dag_id="test")
        with pytest.raises(
            AirflowException,
            match=r"Both asynchronous and deferrable parameters were passed. Please, provide only one.",
        ):
            op.execute(context=mock.MagicMock())

    @pytest.mark.parametrize(
        ("pipeline_id", "runtime_args", "expected_run_id", "expected_runtime_args", "expected_output_suffix"),
        [
            ("abc123", {"arg1": "val1"}, "abc123", {"arg1": "val1"}, "abc123"),
            (None, None, None, None, "unknown"),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.operators.datafusion.DataFusionPipelineLink.persist")
    @mock.patch(HOOK_STR)
    def test_openlineage_facets_with_mock(
        self,
        mock_hook,
        mock_persist,
        pipeline_id,
        runtime_args,
        expected_run_id,
        expected_runtime_args,
        expected_output_suffix,
    ):
        mock_persist.return_value = None

        mock_instance = {"apiEndpoint": "https://mock-endpoint", "serviceEndpoint": "https://mock-service"}
        mock_hook.return_value.get_instance.return_value = mock_instance
        mock_hook.return_value.start_pipeline.return_value = pipeline_id

        op = CloudDataFusionStartPipelineOperator(
            task_id=TASK_ID,
            pipeline_name=PIPELINE_NAME,
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
            runtime_args=runtime_args,
        )

        result_pipeline_id = op.execute(context={})
        results = op.get_openlineage_facets_on_complete(task_instance=None)

        assert result_pipeline_id == pipeline_id
        assert op.pipeline_id == pipeline_id

        expected_input_name = f"{PROJECT_ID}:{LOCATION}:{INSTANCE_NAME}:{PIPELINE_NAME}"

        assert results is not None
        assert len(results.inputs) == 1
        assert results.inputs[0].namespace == "datafusion"
        assert results.inputs[0].name == expected_input_name

        assert len(results.outputs) == 1
        assert results.outputs[0].namespace == "datafusion"
        assert results.outputs[0].name == f"{expected_input_name}:{expected_output_suffix}"

        facet = results.run_facets["dataFusionRun"]
        assert isinstance(facet, DataFusionRunFacet)
        assert facet.runId == expected_run_id
        assert facet.runtimeArgs == expected_runtime_args

        assert results.job_facets == {}


class TestCloudDataFusionStopPipelineOperator:
    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_should_execute_successfully(self, mock_hook):
        mock_hook.return_value.get_instance.return_value = {
            "apiEndpoint": INSTANCE_URL,
            "serviceEndpoint": INSTANCE_URL,
        }
        op = CloudDataFusionStopPipelineOperator(
            task_id="test_tasks",
            pipeline_name=PIPELINE_NAME,
            instance_name=INSTANCE_NAME,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.return_value.get_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )

        mock_hook.return_value.stop_pipeline.assert_called_once_with(
            instance_url=INSTANCE_URL, pipeline_name=PIPELINE_NAME, namespace=NAMESPACE
        )


class TestCloudDataFusionListPipelinesOperator:
    @mock.patch(HOOK_STR)
    def test_execute_check_hook_call_should_execute_successfully(self, mock_hook):
        artifact_version = "artifact_version"
        artifact_name = "artifact_name"
        mock_hook.return_value.get_instance.return_value = {
            "apiEndpoint": INSTANCE_URL,
            "serviceEndpoint": INSTANCE_URL,
        }
        op = CloudDataFusionListPipelinesOperator(
            task_id="test_tasks",
            instance_name=INSTANCE_NAME,
            artifact_version=artifact_version,
            artifact_name=artifact_name,
            namespace=NAMESPACE,
            location=LOCATION,
            project_id=PROJECT_ID,
        )
        op.execute(context=mock.MagicMock())
        mock_hook.return_value.get_instance.assert_called_once_with(
            instance_name=INSTANCE_NAME, location=LOCATION, project_id=PROJECT_ID
        )

        mock_hook.return_value.list_pipelines.assert_called_once_with(
            instance_url=INSTANCE_URL,
            namespace=NAMESPACE,
            artifact_version=artifact_version,
            artifact_name=artifact_name,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that translates text in Google Cloud Translate using V3 API version
service in the Google Cloud.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.operators.translate import (
    TranslateTextBatchOperator,
    TranslateTextOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

DAG_ID = "gcp_translate_text"
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
REGION = "us-central1"
RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"
BATCH_TRANSLATE_SAMPLE_URI = (
    f"gs://{RESOURCE_DATA_BUCKET}/translate/V3/text_batch/inputs/translate_sample_de_1.txt"
)
BATCH_TRANSLATE_INPUT = {
    "gcs_source": {"input_uri": BATCH_TRANSLATE_SAMPLE_URI},
    "mime_type": "text/plain",
}
DATA_SAMPLE_GCS_BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}".replace("_", "-")
GCS_OUTPUT_DST = {
    "gcs_destination": {"output_uri_prefix": f"gs://{DATA_SAMPLE_GCS_BUCKET_NAME}/translate_output/"}
}


with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2024, 11, 1),
    catchup=False,
    tags=["example", "translate_text", "batch_translate_text"],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=DATA_SAMPLE_GCS_BUCKET_NAME,
        storage_class="REGIONAL",
        location=REGION,
    )
    # [START howto_operator_translate_text_advanced]
    translate_text = TranslateTextOperator(
        task_id="translate_v3_op",
        contents=["Ciao mondo!", "Mi puoi prendere una tazza di caffè, per favore?"],
        source_language_code="it",
        target_language_code="en",
    )
    # [END howto_operator_translate_text_advanced]

    # [START howto_operator_batch_translate_text]
    batch_text_translate = TranslateTextBatchOperator(
        task_id="batch_translate_v3_op",
        source_language_code="de",
        target_language_codes=["en"],  # Up to 10 language codes per run
        location="us-central1",
        input_configs=[BATCH_TRANSLATE_INPUT],
        output_config=GCS_OUTPUT_DST,
    )
    # [END howto_operator_batch_translate_text]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=DATA_SAMPLE_GCS_BUCKET_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        create_bucket
        # TEST BODY
        >> [translate_text, batch_text_translate]
        # TEST TEARDOWN
        >> delete_bucket
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from copy import deepcopy
from typing import Any
from unittest import mock

import pytest

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.sensors.dataproc_metastore import MetastoreHivePartitionSensor

DATAPROC_METASTORE_SENSOR_PATH = "airflow.providers.google.cloud.sensors.dataproc_metastore.{}"
TEST_TASK_ID = "test-task"
PARTITION_1 = "column=value"
PARTITION_2 = "column1=value1/column2=value2"
RESULT_FILE_NAME_1 = "result-001"
RESULT_FILE_NAME_2 = "result-002"
MANIFEST_SUCCESS = {
    "status": {
        "code": 0,
        "message": "Query results are successfully uploaded to cloud storage",
        "details": [],
    },
    "filenames": [],
}
MANIFEST_FAIL = {"status": {"code": 1, "message": "Bad things happened", "details": []}, "filenames": []}
RESULT_FILE_CONTENT: dict[str, Any] = {"rows": [], "metadata": {}}
ROW_1: list[Any] = []
ROW_2: list[Any] = []
TEST_SERVICE_ID = "test-service"
TEST_REGION = "test-region"
TEST_TABLE = "test_table"
GCP_PROJECT = "test-project"
GCP_CONN_ID = "test-conn"
TEST_URI = "test-uri"


class TestMetastoreHivePartitionSensor:
    @pytest.mark.parametrize(
        ("requested_partitions", "result_files_with_rows", "expected_result"),
        [
            (None, [(RESULT_FILE_NAME_1, [])], False),
            ([None], [(RESULT_FILE_NAME_1, [])], False),
            (None, [(RESULT_FILE_NAME_1, [ROW_1])], True),
            ([None], [(RESULT_FILE_NAME_1, [ROW_1])], True),
            (None, [(RESULT_FILE_NAME_1, [ROW_1, ROW_2])], True),
            ([None], [(RESULT_FILE_NAME_1, [ROW_1, ROW_2])], True),
            ([PARTITION_1], [(RESULT_FILE_NAME_1, [])], False),
            ([PARTITION_1], [(RESULT_FILE_NAME_1, [ROW_1])], True),
            ([PARTITION_1, PARTITION_2], [(RESULT_FILE_NAME_1, [])], False),
            ([PARTITION_1, PARTITION_2], [(RESULT_FILE_NAME_1, [ROW_1])], False),
            ([PARTITION_1, PARTITION_2], [(RESULT_FILE_NAME_1, [ROW_1, ROW_2])], True),
            ([PARTITION_1, PARTITION_1], [(RESULT_FILE_NAME_1, [])], False),
            ([PARTITION_1, PARTITION_1], [(RESULT_FILE_NAME_1, [ROW_1])], True),
            ([PARTITION_1, PARTITION_2], [(RESULT_FILE_NAME_1, []), (RESULT_FILE_NAME_2, [])], False),
            ([PARTITION_1, PARTITION_2], [(RESULT_FILE_NAME_1, [ROW_1]), (RESULT_FILE_NAME_2, [])], False),
            ([PARTITION_1, PARTITION_2], [(RESULT_FILE_NAME_1, []), (RESULT_FILE_NAME_2, [ROW_2])], False),
            (
                [PARTITION_1, PARTITION_2],
                [(RESULT_FILE_NAME_1, [ROW_1]), (RESULT_FILE_NAME_2, [ROW_2])],
                True,
            ),
        ],
    )
    @mock.patch(DATAPROC_METASTORE_SENSOR_PATH.format("DataprocMetastoreHook"))
    @mock.patch(DATAPROC_METASTORE_SENSOR_PATH.format("parse_json_from_gcs"))
    def test_poke_positive_manifest(
        self,
        mock_parse_json_from_gcs,
        mock_hook,
        requested_partitions,
        result_files_with_rows,
        expected_result,
    ):
        manifest = deepcopy(MANIFEST_SUCCESS)
        parse_json_from_gcs_side_effect = []
        for file_name, rows in result_files_with_rows:
            manifest["filenames"].append(file_name)
            file = deepcopy(RESULT_FILE_CONTENT)
            file["rows"] = rows
            parse_json_from_gcs_side_effect.append(file)

        mock_parse_json_from_gcs.side_effect = [manifest, *parse_json_from_gcs_side_effect]

        sensor = MetastoreHivePartitionSensor(
            task_id=TEST_TASK_ID,
            service_id=TEST_SERVICE_ID,
            region=TEST_REGION,
            table=TEST_TABLE,
            partitions=requested_partitions,
            gcp_conn_id=GCP_CONN_ID,
        )
        assert sensor.poke(context={}) == expected_result

    @pytest.mark.parametrize("empty_manifest", [dict(), list(), tuple(), None, ""])
    @mock.patch(DATAPROC_METASTORE_SENSOR_PATH.format("DataprocMetastoreHook"))
    @mock.patch(DATAPROC_METASTORE_SENSOR_PATH.format("parse_json_from_gcs"))
    def test_poke_empty_manifest(self, mock_parse_json_from_gcs, mock_hook, empty_manifest):
        mock_parse_json_from_gcs.return_value = empty_manifest

        sensor = MetastoreHivePartitionSensor(
            task_id=TEST_TASK_ID,
            service_id=TEST_SERVICE_ID,
            region=TEST_REGION,
            table=TEST_TABLE,
            partitions=[PARTITION_1],
            gcp_conn_id=GCP_CONN_ID,
        )

        with pytest.raises(AirflowException):
            sensor.poke(context={})

    @mock.patch(DATAPROC_METASTORE_SENSOR_PATH.format("DataprocMetastoreHook"))
    @mock.patch(DATAPROC_METASTORE_SENSOR_PATH.format("parse_json_from_gcs"))
    def test_poke_wrong_status(self, mock_parse_json_from_gcs, mock_hook):
        error_message = "Test error message"
        mock_parse_json_from_gcs.return_value = {"code": 1, "message": error_message}

        sensor = MetastoreHivePartitionSensor(
            task_id=TEST_TASK_ID,
            service_id=TEST_SERVICE_ID,
            region=TEST_REGION,
            table=TEST_TABLE,
            partitions=[PARTITION_1],
            gcp_conn_id=GCP_CONN_ID,
        )

        with pytest.raises(AirflowException, match=f"Request failed: {error_message}"):
            sensor.poke(context={})

    @pytest.mark.parametrize(
        ("requested_partitions", "result_files_with_rows", "expected_result"),
        [
            ([PARTITION_1, PARTITION_1], [(RESULT_FILE_NAME_1, [ROW_1])], True),
        ],
    )
    @mock.patch(DATAPROC_METASTORE_SENSOR_PATH.format("DataprocMetastoreHook"))
    @mock.patch(DATAPROC_METASTORE_SENSOR_PATH.format("parse_json_from_gcs"))
    def test_file_uri(
        self,
        mock_parse_json_from_gcs,
        mock_hook,
        requested_partitions,
        result_files_with_rows,
        expected_result,
    ):
        mock_hook.return_value.wait_for_operation.return_value = mock.MagicMock(result_manifest_uri=TEST_URI)
        manifest = deepcopy(MANIFEST_SUCCESS)
        parse_json_from_gcs_side_effect = []
        for file_name, rows in result_files_with_rows:
            manifest["filenames"].append(file_name)
            file = deepcopy(RESULT_FILE_CONTENT)
            file["rows"] = rows
            parse_json_from_gcs_side_effect.append(file)

        mock_parse_json_from_gcs.side_effect = [manifest, *parse_json_from_gcs_side_effect]

        sensor = MetastoreHivePartitionSensor(
            task_id=TEST_TASK_ID,
            service_id=TEST_SERVICE_ID,
            region=TEST_REGION,
            table=TEST_TABLE,
            partitions=requested_partitions,
            gcp_conn_id=GCP_CONN_ID,
        )
        assert sensor.poke(context={}) == expected_result
        mock_parse_json_from_gcs.assert_called_with(
            file_uri=TEST_URI + "/" + RESULT_FILE_NAME_1, gcp_conn_id=GCP_CONN_ID, impersonation_chain=None
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.suite.operators.sheets import GoogleSheetsCreateSpreadsheetOperator

GCP_CONN_ID = "test"
SPREADSHEET_URL = "https://example/sheets"
SPREADSHEET_ID = "1234567890"


class TestGoogleSheetsCreateSpreadsheet:
    @mock.patch("airflow.providers.google.suite.operators.sheets.GSheetsHook")
    def test_execute(self, mock_hook):
        mock_task_instance = mock.MagicMock()
        context = {"task_instance": mock_task_instance}
        spreadsheet = mock.MagicMock()
        mock_hook.return_value.create_spreadsheet.return_value = {
            "spreadsheetId": SPREADSHEET_ID,
            "spreadsheetUrl": SPREADSHEET_URL,
        }
        op = GoogleSheetsCreateSpreadsheetOperator(
            task_id="test_task", spreadsheet=spreadsheet, gcp_conn_id=GCP_CONN_ID
        )
        op_execute_result = op.execute(context)

        mock_hook.return_value.create_spreadsheet.assert_called_once_with(spreadsheet=spreadsheet)

        # Verify xcom_push was called with correct arguments
        assert mock_task_instance.xcom_push.call_count == 2
        mock_task_instance.xcom_push.assert_any_call(key="spreadsheet_id", value=SPREADSHEET_ID)
        mock_task_instance.xcom_push.assert_any_call(key="spreadsheet_url", value=SPREADSHEET_URL)

        assert op_execute_result["spreadsheetId"] == "1234567890"
        assert op_execute_result["spreadsheetUrl"] == "https://example/sheets"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from types import SimpleNamespace
from unittest import mock
from unittest.mock import AsyncMock

import pytest
from google.cloud.storage_transfer_v1.types.transfer_types import TransferOperation

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.hooks.cloud_storage_transfer_service import (
    CloudDataTransferServiceAsyncHook,
    GcpTransferOperationStatus,
)

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

TEST_PROJECT_ID = "project-id"
TRANSFER_HOOK_PATH = "airflow.providers.google.cloud.hooks.cloud_storage_transfer_service"


@pytest.fixture
def hook_async():
    with mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    ):
        yield CloudDataTransferServiceAsyncHook()


class TestCloudDataTransferServiceAsyncHook:
    @pytest.mark.asyncio
    @mock.patch(f"{TRANSFER_HOOK_PATH}.CloudDataTransferServiceAsyncHook.get_conn")
    @mock.patch(f"{TRANSFER_HOOK_PATH}.StorageTransferServiceAsyncClient")
    async def test_get_conn(self, mock_async_client, mock_get_conn):
        expected_value = "Async Hook"
        mock_async_client.return_value = expected_value
        mock_get_conn.return_value = expected_value

        hook = CloudDataTransferServiceAsyncHook(project_id=TEST_PROJECT_ID)

        conn_0 = await hook.get_conn()
        assert conn_0 == expected_value

        conn_1 = await hook.get_conn()
        assert conn_1 == expected_value
        assert id(conn_0) == id(conn_1)

    @pytest.mark.asyncio
    @mock.patch(f"{TRANSFER_HOOK_PATH}.CloudDataTransferServiceAsyncHook.get_conn")
    @mock.patch(f"{TRANSFER_HOOK_PATH}.ListTransferJobsRequest")
    async def test_get_jobs(self, mock_list_jobs_request, mock_get_conn):
        expected_jobs = AsyncMock()
        mock_get_conn.return_value.list_transfer_jobs.side_effect = AsyncMock(return_value=expected_jobs)

        expected_request = mock.MagicMock()
        mock_list_jobs_request.return_value = expected_request

        hook = CloudDataTransferServiceAsyncHook(project_id=TEST_PROJECT_ID)
        job_names = ["Job0", "Job1"]
        jobs = await hook.get_jobs(job_names=job_names)

        assert jobs == expected_jobs
        mock_list_jobs_request.assert_called_once_with(
            filter=json.dumps(dict(project_id=TEST_PROJECT_ID, job_names=job_names))
        )
        mock_get_conn.return_value.list_transfer_jobs.assert_called_once_with(request=expected_request)

    @pytest.mark.asyncio
    @mock.patch(f"{TRANSFER_HOOK_PATH}.CloudDataTransferServiceAsyncHook.get_conn")
    @mock.patch(f"{TRANSFER_HOOK_PATH}.TransferOperation.deserialize")
    async def test_get_last_operation(self, mock_deserialize, mock_conn, hook_async):
        latest_operation_name = "Mock operation name"
        operation_metadata_value = "Mock metadata value"

        get_operation = AsyncMock()
        get_operation.return_value = mock.MagicMock(metadata=mock.MagicMock(value=operation_metadata_value))
        mock_conn.return_value.transport.operations_client.get_operation = get_operation

        expected_operation = mock.MagicMock()
        mock_deserialize.return_value = expected_operation

        operation = await hook_async.get_latest_operation(
            job=mock.MagicMock(latest_operation_name=latest_operation_name)
        )

        get_operation.assert_called_once_with(latest_operation_name)
        mock_deserialize.assert_called_once_with(operation_metadata_value)
        assert operation == expected_operation

    @pytest.mark.asyncio
    @mock.patch(f"{TRANSFER_HOOK_PATH}.CloudDataTransferServiceAsyncHook.get_conn")
    @mock.patch(f"{TRANSFER_HOOK_PATH}.TransferOperation.deserialize")
    async def test_get_last_operation_none(self, mock_deserialize, mock_conn, hook_async):
        latest_operation_name = None
        expected_operation = None

        get_operation = mock.MagicMock()
        mock_conn.return_value.transport.operations_client.get_operation = get_operation

        operation = await hook_async.get_latest_operation(
            job=mock.MagicMock(latest_operation_name=latest_operation_name)
        )

        get_operation.assert_not_called()
        mock_deserialize.assert_not_called()
        assert operation == expected_operation

    @pytest.mark.asyncio
    @mock.patch(f"{TRANSFER_HOOK_PATH}.CloudDataTransferServiceAsyncHook.get_conn")
    @mock.patch(f"{TRANSFER_HOOK_PATH}.MessageToDict")
    async def test_list_transfer_operations(self, message_to_dict, mock_conn, hook_async):
        expected = [{"name": "op1"}, {"name": "op2"}]
        message_to_dict.side_effect = expected

        op_with_pb = SimpleNamespace(_pb=mock.sentinel.pb1)
        op_without_pb = object()

        first_page = mock.MagicMock(next_page_token="token", operations=[op_with_pb])
        second_page = mock.MagicMock(next_page_token=None, operations=[op_without_pb])
        mock_conn.return_value.list_operations.side_effect = [first_page, second_page]

        actual = await hook_async.list_transfer_operations(
            request_filter={"project_id": TEST_PROJECT_ID},
        )

        assert actual == expected
        assert mock_conn.return_value.list_operations.call_count == 2
        assert message_to_dict.call_args_list == [
            mock.call(mock.sentinel.pb1, preserving_proto_field_name=True, use_integers_for_enums=True),
            mock.call(op_without_pb, preserving_proto_field_name=True, use_integers_for_enums=True),
        ]

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ("statuses", "expected_statuses"),
        [
            ([GcpTransferOperationStatus.ABORTED], (GcpTransferOperationStatus.IN_PROGRESS,)),
            (
                [GcpTransferOperationStatus.SUCCESS, GcpTransferOperationStatus.ABORTED],
                (GcpTransferOperationStatus.IN_PROGRESS,),
            ),
            (
                [GcpTransferOperationStatus.PAUSED, GcpTransferOperationStatus.ABORTED],
                (GcpTransferOperationStatus.IN_PROGRESS,),
            ),
        ],
    )
    async def test_operations_contain_expected_statuses_red_path(self, statuses, expected_statuses):
        def to_name(x):
            return x.name if hasattr(x, "name") else x

        def proto_int(name: str) -> int:
            return int(getattr(TransferOperation.Status, name))

        operations = [{"metadata": {"status": proto_int(to_name(s))}} for s in statuses]

        expected_names = tuple(to_name(s) for s in expected_statuses)

        with pytest.raises(
            AirflowException,
            match=f"An unexpected operation status was encountered. Expected: {', '.join(expected_names)}",
        ):
            await CloudDataTransferServiceAsyncHook.operations_contain_expected_statuses(
                operations,
                GcpTransferOperationStatus.IN_PROGRESS,
            )

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ("statuses", "expected_statuses"),
        [
            ([GcpTransferOperationStatus.ABORTED], GcpTransferOperationStatus.ABORTED),
            (
                [GcpTransferOperationStatus.SUCCESS, GcpTransferOperationStatus.ABORTED],
                GcpTransferOperationStatus.ABORTED,
            ),
            (
                [GcpTransferOperationStatus.PAUSED, GcpTransferOperationStatus.ABORTED],
                GcpTransferOperationStatus.ABORTED,
            ),
            ([GcpTransferOperationStatus.ABORTED], (GcpTransferOperationStatus.ABORTED,)),
            (
                [GcpTransferOperationStatus.SUCCESS, GcpTransferOperationStatus.ABORTED],
                (GcpTransferOperationStatus.ABORTED,),
            ),
            (
                [GcpTransferOperationStatus.PAUSED, GcpTransferOperationStatus.ABORTED],
                (GcpTransferOperationStatus.ABORTED,),
            ),
        ],
    )
    async def test_operations_contain_expected_statuses_green_path(self, statuses, expected_statuses):
        to_name = lambda x: x.name if hasattr(x, "name") else x
        name_to_proto_int = lambda name: int(getattr(TransferOperation.Status, name))

        operations = [{"metadata": {"status": name_to_proto_int(to_name(s))}} for s in statuses]

        if isinstance(expected_statuses, (list, tuple, set)):
            expected_norm = {to_name(s) for s in expected_statuses}
        else:
            expected_norm = to_name(expected_statuses)

        result = await CloudDataTransferServiceAsyncHook.operations_contain_expected_statuses(
            operations, expected_norm
        )
        assert result is True

    @pytest.mark.asyncio
    @mock.patch(f"{TRANSFER_HOOK_PATH}.CloudDataTransferServiceAsyncHook.get_conn")
    @mock.patch(f"{TRANSFER_HOOK_PATH}.RunTransferJobRequest")
    async def test_run_transfer_job(self, mock_run_transfer_job_request, mock_get_conn):
        expected_job_result = AsyncMock()
        mock_get_conn.return_value.run_transfer_job.side_effect = AsyncMock(return_value=expected_job_result)

        expected_request = mock.MagicMock()
        mock_run_transfer_job_request.return_value = expected_request

        hook = CloudDataTransferServiceAsyncHook(project_id=TEST_PROJECT_ID)
        job_name = "Job0"
        jobs = await hook.run_transfer_job(job_name=job_name)

        assert jobs == expected_job_result
        mock_run_transfer_job_request.assert_called_once_with(project_id=TEST_PROJECT_ID, job_name=job_name)
        mock_get_conn.return_value.run_transfer_job.assert_called_once_with(request=expected_request)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for Google Cloud Storage to SFTP transfer operators.
"""

from __future__ import annotations

import os
from datetime import datetime
from pathlib import Path

from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.transfers.sftp_to_gcs import SFTPToGCSOperator
from airflow.providers.standard.operators.bash import BashOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
from system.openlineage.operator import OpenLineageTestOperator

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "example_sftp_to_gcs"
BUCKET_NAME = f"bucket-{DAG_ID}-{ENV_ID}"

TMP_PATH = "tmp"
DIR = "tests_sftp_hook_dir"
SUBDIR = "subdir"

OBJECT_SRC_1 = "parent-1.bin"
OBJECT_SRC_2 = "parent-2.bin"

CURRENT_FOLDER = Path(__file__).parent
LOCAL_PATH = str(Path(CURRENT_FOLDER) / "resources")

FILE_LOCAL_PATH = str(Path(LOCAL_PATH) / TMP_PATH / DIR)
FILE_NAME = "tmp.tar.gz"


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    create_bucket = GCSCreateBucketOperator(task_id="create_bucket", bucket_name=BUCKET_NAME)

    unzip_file = BashOperator(
        task_id="unzip_data_file", bash_command=f"tar xvf {LOCAL_PATH}/{FILE_NAME} -C {LOCAL_PATH}"
    )

    # [START howto_operator_sftp_to_gcs_copy_single_file]
    copy_file_from_sftp_to_gcs = SFTPToGCSOperator(
        task_id="file-copy-sftp-to-gcs",
        source_path=f"{FILE_LOCAL_PATH}/{OBJECT_SRC_1}",
        destination_bucket=BUCKET_NAME,
    )
    # [END howto_operator_sftp_to_gcs_copy_single_file]

    # [START howto_operator_sftp_to_gcs_move_single_file_destination]
    move_file_from_sftp_to_gcs_destination = SFTPToGCSOperator(
        task_id="file-move-sftp-to-gcs-destination",
        source_path=f"{FILE_LOCAL_PATH}/{OBJECT_SRC_2}",
        destination_bucket=BUCKET_NAME,
        destination_path="destination_dir/destination_filename.bin",
        move_object=True,
    )
    # [END howto_operator_sftp_to_gcs_move_single_file_destination]

    # [START howto_operator_sftp_to_gcs_copy_directory]
    copy_directory_from_sftp_to_gcs = SFTPToGCSOperator(
        task_id="dir-copy-sftp-to-gcs",
        source_path=f"{FILE_LOCAL_PATH}/{SUBDIR}/*",
        destination_bucket=BUCKET_NAME,
    )
    # [END howto_operator_sftp_to_gcs_copy_directory]

    # [START howto_operator_sftp_to_gcs_move_specific_files]
    move_specific_files_from_sftp_to_gcs = SFTPToGCSOperator(
        task_id="dir-move-specific-files-sftp-to-gcs",
        source_path=f"{FILE_LOCAL_PATH}/{SUBDIR}/*.bin",
        destination_bucket=BUCKET_NAME,
        destination_path="specific_files/",
        move_object=True,
    )
    # [END howto_operator_sftp_to_gcs_move_specific_files]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    check_openlineage_events = OpenLineageTestOperator(
        task_id="check_openlineage_events",
        file_path=str(Path(__file__).parent / "resources" / "openlineage" / "sftp_to_gcs.json"),
    )

    chain(
        # TEST SETUP
        create_bucket,
        unzip_file,
        # TEST BODY
        copy_file_from_sftp_to_gcs,
        move_file_from_sftp_to_gcs_destination,
        copy_directory_from_sftp_to_gcs,
        move_specific_files_from_sftp_to_gcs,
        # TEST TEARDOWN
        delete_bucket,
        check_openlineage_events,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import copy
import io
import logging
import os
from types import GeneratorType
from typing import TYPE_CHECKING
from unittest import mock
from unittest.mock import MagicMock

import pytest

from airflow.providers.google.cloud.log.gcs_task_handler import GCSRemoteLogIO, GCSTaskHandler
from airflow.utils.state import TaskInstanceState
from airflow.utils.timezone import datetime

from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.db import clear_db_dags, clear_db_runs
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if TYPE_CHECKING:
    from pathlib import Path


def patch_mock_client_for_list_blobs(mock_client: MagicMock, blob_names: list[str]):
    mock_blobs = []
    for name in blob_names:
        mock_blob = MagicMock()
        mock_blob.name = name
        mock_blobs.append(mock_blob)
    mock_client.return_value.list_blobs.return_value = mock_blobs


@pytest.mark.db_test
@mock.patch(
    "airflow.providers.google.cloud.log.gcs_task_handler.get_credentials_and_project_id",
    return_value=("TEST_CREDENTIALS", "TEST_PROJECT_ID"),
)
class TestGCSRemoteLogIO:
    @pytest.fixture(autouse=True)
    def setup_tests(self, create_task_instance, session):
        # setup remote IO
        self.base_log_folder = "local/airflow/logs"
        self.gcs_log_folder = "gs://bucket/airflow/logs"
        # setup TaskInstance
        self.ti = ti = create_task_instance(task_id="task_1")
        ti.try_number = 1
        ti.raw = False
        session.add(ti)
        session.commit()
        yield
        clear_db_runs()
        clear_db_dags()

    @pytest.mark.parametrize(
        "is_absolute",
        [pytest.param(True, id="absolute"), pytest.param(False, id="relative")],
    )
    @pytest.mark.parametrize(
        "file_exists",
        [pytest.param(True, id="file-exists"), pytest.param(False, id="file-not-exists")],
    )
    @pytest.mark.parametrize(
        "delete_local_copy",
        [pytest.param(True, id="delete-local"), pytest.param(False, id="keep-local")],
    )
    @pytest.mark.parametrize(
        "mock_write_method_result",
        [pytest.param(True, id="write-success"), pytest.param(False, id="write-fail")],
    )
    @mock.patch("google.cloud.storage.Client")
    @mock.patch("google.cloud.storage.Blob")
    @mock.patch("shutil.rmtree")
    def test_upload(
        self,
        mock_rmtree,
        mock_blob,
        mock_client,
        mock_creds,
        mock_write_method_result: bool,
        delete_local_copy: bool,
        file_exists: bool,
        is_absolute: bool,
        tmp_path: Path,
    ):
        # setup
        gcs_remote_log_io = GCSRemoteLogIO(
            remote_base=self.gcs_log_folder,
            base_log_folder=tmp_path.as_posix(),
            delete_local_copy=delete_local_copy,
        )
        if file_exists:
            file_path = tmp_path / "existing.log" if is_absolute else "existing.log"
            with open(tmp_path / "existing.log", "w") as f:
                f.write("log content")
        else:
            file_path = tmp_path / "non_existing.log"

        # action
        with mock.patch.object(
            gcs_remote_log_io,
            "write",
            return_value=mock_write_method_result,
        ) as mock_write_method:
            gcs_remote_log_io.upload(file_path, self.ti)

            # verify
            if file_exists:
                mock_write_method.assert_called_once()
                if delete_local_copy and mock_write_method_result:
                    mock_rmtree.assert_called_once_with(tmp_path.as_posix())
                else:
                    mock_rmtree.assert_not_called()
            else:
                mock_write_method.assert_not_called()
                mock_rmtree.assert_not_called()

    @pytest.mark.parametrize(
        "upload_success",
        [pytest.param(True, id="upload-success"), pytest.param(False, id="upload-fail")],
    )
    @pytest.mark.parametrize(
        "old_log_exists",
        [
            pytest.param(True, id="old-log-exists"),
            pytest.param(False, id="old-log-not-exists"),
        ],
    )
    @pytest.mark.parametrize(
        "old_log_read_error",
        [
            pytest.param(None, id="old-log-read-success"),
            pytest.param(Exception("Read error"), id="old-log-read-error"),
        ],
    )
    @mock.patch("google.cloud.storage.Client")
    @mock.patch("google.cloud.storage.Blob")
    def test_write(
        self,
        mock_blob,
        mock_client,
        mock_creds,
        upload_success: bool,
        old_log_exists: bool,
        old_log_read_error: Exception | None,
    ):
        # setup
        remote_log_location = f"{self.gcs_log_folder}/task_1/attempt_1.log"
        new_log_content = "NEW LOG CONTENT"
        old_log_content = "OLD LOG CONTENT"

        # mock download_as_bytes for reading old log
        if old_log_read_error:
            mock_blob.from_string.return_value.download_as_bytes.side_effect = old_log_read_error
        elif old_log_exists:
            mock_blob.from_string.return_value.download_as_bytes.return_value = old_log_content.encode()
        else:
            # Simulate 404 error for no log found
            not_found_error = Exception("No such object: bucket/airflow/logs/task_1/attempt_1.log")
            mock_blob.from_string.return_value.download_as_bytes.side_effect = not_found_error

        # mock upload_from_string
        if not upload_success:
            mock_blob.from_string.return_value.upload_from_string.side_effect = Exception("Upload failed")

        gcs_remote_log_io = GCSRemoteLogIO(
            remote_base=self.gcs_log_folder,
            base_log_folder=self.base_log_folder,
            delete_local_copy=False,
        )

        # action
        result = gcs_remote_log_io.write(new_log_content, remote_log_location)

        # verify
        assert result == upload_success

        # verify the content that was uploaded
        if upload_success:
            call_args = mock_blob.from_string.return_value.upload_from_string.call_args
            if call_args:
                uploaded_content = call_args[0][0]
                if old_log_exists and not old_log_read_error:
                    assert uploaded_content == f"{old_log_content}\n{new_log_content}"
                else:
                    assert uploaded_content == new_log_content

    @pytest.mark.parametrize(
        "is_stream_method",
        [pytest.param(True, id="is-stream"), pytest.param(False, id="not-stream")],
    )
    @pytest.mark.parametrize(
        "blob_names",
        [
            pytest.param(
                ["airflow/logs/task_1/attempt_1.log", "airflow/logs/task_1/attempt_2.log"], id="blobs-exists"
            ),
            pytest.param([], id="blobs-not-exists"),
        ],
    )
    @pytest.mark.parametrize(
        "read_success",
        [pytest.param(True, id="read-success"), pytest.param(False, id="read-fail")],
    )
    @mock.patch("google.cloud.storage.Client")
    @mock.patch("google.cloud.storage.Blob")
    def test_stream_and_read_methods(
        self,
        mock_blob,
        mock_client,
        mock_creds,
        read_success: bool,
        blob_names: list[str],
        is_stream_method: bool,
    ):
        # setup
        patch_mock_client_for_list_blobs(mock_client, blob_names)
        if read_success:
            mock_blob.from_string.return_value.open.side_effect = lambda mode: io.TextIOWrapper(
                io.BytesIO(b"LOG\nCONTENT"), encoding="utf-8"
            )
        else:
            mock_blob.from_string.return_value.open.side_effect = Exception("Read failed")

        gcs_remote_log_io = GCSRemoteLogIO(
            remote_base=self.gcs_log_folder,
            base_log_folder=self.base_log_folder,
            delete_local_copy=False,
        )
        # action
        if is_stream_method:
            messages, log_streams = gcs_remote_log_io.stream("airflow/logs/task_1", self.ti)
            logs = log_streams  # type: ignore[assignment]
        else:
            messages, logs = gcs_remote_log_io.read("airflow/logs/task_1", self.ti)  # type: ignore[assignment]

        # early return for no blobs
        if not blob_names:
            assert messages == []
            if is_stream_method:
                assert logs == []
            else:
                assert logs is None
            return

        # verify messages
        expected_uris = [
            f"{self.gcs_log_folder}/task_1/attempt_1.log",
            f"{self.gcs_log_folder}/task_1/attempt_2.log",
        ]
        if AIRFLOW_V_3_0_PLUS:
            expected_messages = expected_uris
        else:
            expected_messages = ["Found remote logs:", *[f"  * {x}" for x in sorted(expected_uris)]]
        if not read_success and not AIRFLOW_V_3_0_PLUS:
            expected_messages = expected_messages + ["Unable to read remote log Read failed"]
        assert messages == expected_messages

        # verify logs
        expected_logs = ["LOG\nCONTENT", "LOG\nCONTENT"]
        if is_stream_method:
            for log_stream, expected_log in zip(logs, expected_logs):
                assert isinstance(log_stream, GeneratorType)
                assert "".join(log_stream) == expected_log
        else:
            if read_success:
                assert logs == expected_logs
            else:
                assert logs is None


@pytest.mark.db_test
class TestGCSTaskHandler:
    @pytest.fixture(autouse=True)
    def task_instance(self, create_task_instance, session):
        self.ti = ti = create_task_instance(
            dag_id="dag_for_testing_gcs_task_handler",
            task_id="task_for_testing_gcs_task_handler",
            logical_date=datetime(2020, 1, 1),
            state=TaskInstanceState.RUNNING,
        )
        ti.try_number = 1
        ti.raw = False
        session.add(ti)
        session.commit()
        yield
        clear_db_runs()
        clear_db_dags()

    @pytest.fixture(autouse=True)
    def local_log_location(self, tmp_path_factory):
        return str(tmp_path_factory.mktemp("local-gcs-log-location"))

    @pytest.fixture(autouse=True)
    def gcs_task_handler(self, create_log_template, local_log_location):
        create_log_template("{try_number}.log")
        self.gcs_task_handler = GCSTaskHandler(
            base_log_folder=local_log_location,
            gcs_log_folder="gs://bucket/remote/log/location",
        )
        return self.gcs_task_handler

    @mock.patch("airflow.providers.google.cloud.log.gcs_task_handler.GCSHook")
    @mock.patch("google.cloud.storage.Client")
    @mock.patch("airflow.providers.google.cloud.log.gcs_task_handler.get_credentials_and_project_id")
    @pytest.mark.parametrize(
        "conn_id",
        [pytest.param("", id="no-conn"), pytest.param("my_gcs_conn", id="with-conn")],
    )
    def test_client_conn_id_behavior(self, mock_get_cred, mock_client, mock_hook, conn_id):
        """When remote log conn id configured, hook will be used"""
        mock_hook.return_value.get_credentials_and_project_id.return_value = (
            "test_cred",
            "test_proj",
        )
        mock_get_cred.return_value = ("test_cred", "test_proj")
        with conf_vars({("logging", "remote_log_conn_id"): conn_id}):
            return_value = self.gcs_task_handler.io.client
        if conn_id:
            mock_hook.assert_called_once_with(gcp_conn_id="my_gcs_conn")
            mock_get_cred.assert_not_called()
        else:
            mock_hook.assert_not_called()
            mock_get_cred.assert_called()

        mock_client.assert_called_once_with(
            client_info=mock.ANY, credentials="test_cred", project="test_proj"
        )
        assert mock_client.return_value == return_value

    @conf_vars({("logging", "remote_log_conn_id"): "gcs_default"})
    @mock.patch(
        "airflow.providers.google.cloud.log.gcs_task_handler.get_credentials_and_project_id",
        return_value=("TEST_CREDENTIALS", "TEST_PROJECT_ID"),
    )
    @mock.patch("google.cloud.storage.Client")
    @mock.patch("google.cloud.storage.Blob")
    def test_should_read_logs_from_remote(
        self, mock_blob, mock_client, mock_creds, session, sdk_connection_not_found
    ):
        blob_name = "remote/log/location/1.log"
        patch_mock_client_for_list_blobs(mock_client, [blob_name])
        mock_blob.from_string.return_value.open.return_value = io.TextIOWrapper(
            io.BytesIO(b"CONTENT"), encoding="utf-8"
        )
        ti = copy.copy(self.ti)
        ti.state = TaskInstanceState.SUCCESS
        session.add(ti)
        session.commit()
        logs, metadata = self.gcs_task_handler._read(ti, self.ti.try_number)
        expected_gs_uri = f"gs://bucket/{blob_name}"

        mock_blob.from_string.assert_called_once_with(expected_gs_uri, mock_client.return_value)

        if AIRFLOW_V_3_0_PLUS:
            logs = list(logs)
            assert logs[0].event == "::group::Log message source details"
            assert logs[0].sources == [expected_gs_uri]
            assert logs[1].event == "::endgroup::"
            assert logs[2].event == "CONTENT"
            assert metadata == {"end_of_log": True, "log_pos": 1}
        else:
            assert f"*** Found remote logs:\n***   * {expected_gs_uri}\n" in logs
            assert logs.endswith("CONTENT")
            assert metadata == {"end_of_log": True, "log_pos": 7}

    @mock.patch(
        "airflow.providers.google.cloud.log.gcs_task_handler.get_credentials_and_project_id",
        return_value=("TEST_CREDENTIALS", "TEST_PROJECT_ID"),
    )
    @mock.patch("google.cloud.storage.Client")
    @mock.patch("google.cloud.storage.Blob")
    def test_should_read_from_local_on_logs_read_error(self, mock_blob, mock_client, mock_creds):
        blob_name = "remote/log/location/1.log"
        patch_mock_client_for_list_blobs(mock_client, [blob_name])
        mock_blob.from_string.return_value.open.side_effect = Exception("Failed to connect")

        self.gcs_task_handler.set_context(self.ti)
        ti = copy.copy(self.ti)
        ti.state = TaskInstanceState.SUCCESS
        log, metadata = self.gcs_task_handler._read(ti, self.ti.try_number)
        expected_gs_uri = f"gs://bucket/{blob_name}"

        if AIRFLOW_V_3_0_PLUS:
            log = list(log)
            assert log[0].event == "::group::Log message source details"
            assert log[0].sources == [
                expected_gs_uri,
                f"{self.gcs_task_handler.local_base}/1.log",
            ]
            assert log[1].event == "::endgroup::"
            assert metadata == {"end_of_log": True, "log_pos": 0}
        else:
            assert (
                "*** Found remote logs:\n"
                "***   * gs://bucket/remote/log/location/1.log\n"
                "*** Unable to read remote log Failed to connect\n"
                "*** Found local files:\n"
                f"***   * {self.gcs_task_handler.local_base}/1.log\n"
            ) in log
            assert metadata == {"end_of_log": True, "log_pos": 0}
        mock_blob.from_string.assert_called_once_with(expected_gs_uri, mock_client.return_value)

    @mock.patch(
        "airflow.providers.google.cloud.log.gcs_task_handler.get_credentials_and_project_id",
        return_value=("TEST_CREDENTIALS", "TEST_PROJECT_ID"),
    )
    @mock.patch("google.cloud.storage.Client")
    @mock.patch("google.cloud.storage.Blob")
    def test_write_to_remote_on_close(self, mock_blob, mock_client, mock_creds):
        mock_blob.from_string.return_value.download_as_bytes.return_value = b"CONTENT"

        self.gcs_task_handler.set_context(self.ti)
        self.gcs_task_handler.emit(
            logging.LogRecord(
                name="NAME",
                level="DEBUG",
                pathname=None,
                lineno=None,
                msg="MESSAGE",
                args=None,
                exc_info=None,
            )
        )
        self.gcs_task_handler.close()

        mock_blob.assert_has_calls(
            [
                mock.call.from_string("gs://bucket/remote/log/location/1.log", mock_client.return_value),
                mock.call.from_string().download_as_bytes(),
                mock.call.from_string("gs://bucket/remote/log/location/1.log", mock_client.return_value),
                mock.call.from_string().upload_from_string("CONTENT\nMESSAGE\n", content_type="text/plain"),
            ],
            any_order=False,
        )
        mock_blob.from_string.return_value.upload_from_string(data="CONTENT\nMESSAGE\n")
        assert self.gcs_task_handler.closed is True

    @mock.patch(
        "airflow.providers.google.cloud.log.gcs_task_handler.get_credentials_and_project_id",
        return_value=("TEST_CREDENTIALS", "TEST_PROJECT_ID"),
    )
    @mock.patch("google.cloud.storage.Client")
    @mock.patch("google.cloud.storage.Blob")
    def test_failed_write_to_remote_on_close(self, mock_blob, mock_client, mock_creds, caplog):
        caplog.at_level(logging.ERROR, logger=self.gcs_task_handler.log.name)
        mock_blob.from_string.return_value.upload_from_string.side_effect = Exception("Failed to connect")
        mock_blob.from_string.return_value.download_as_bytes.return_value = b"Old log"

        self.gcs_task_handler.set_context(self.ti)
        self.gcs_task_handler.emit(
            logging.LogRecord(
                name="NAME",
                level="DEBUG",
                pathname=None,
                lineno=None,
                msg="MESSAGE",
                args=None,
                exc_info=None,
            )
        )
        self.gcs_task_handler.close()

        assert caplog.record_tuples == [
            (
                "airflow.providers.google.cloud.log.gcs_task_handler.GCSRemoteLogIO",
                logging.ERROR,
                "Could not write logs to gs://bucket/remote/log/location/1.log: Failed to connect",
            ),
        ]
        mock_blob.assert_has_calls(
            [
                mock.call.from_string("gs://bucket/remote/log/location/1.log", mock_client.return_value),
                mock.call.from_string().download_as_bytes(),
                mock.call.from_string("gs://bucket/remote/log/location/1.log", mock_client.return_value),
                mock.call.from_string().upload_from_string("Old log\nMESSAGE\n", content_type="text/plain"),
            ],
            any_order=False,
        )

    @mock.patch(
        "airflow.providers.google.cloud.log.gcs_task_handler.get_credentials_and_project_id",
        return_value=("TEST_CREDENTIALS", "TEST_PROJECT_ID"),
    )
    @mock.patch("google.cloud.storage.Client")
    @mock.patch("google.cloud.storage.Blob")
    def test_write_to_remote_on_close_failed_read_old_logs(self, mock_blob, mock_client, mock_creds):
        mock_blob.from_string.return_value.download_as_bytes.side_effect = Exception("Fail to download")

        self.gcs_task_handler.set_context(self.ti)
        self.gcs_task_handler.emit(
            logging.LogRecord(
                name="NAME",
                level="DEBUG",
                pathname=None,
                lineno=None,
                msg="MESSAGE",
                args=None,
                exc_info=None,
            )
        )
        self.gcs_task_handler.close()

        mock_blob.from_string.assert_has_calls(
            [
                mock.call("gs://bucket/remote/log/location/1.log", mock_client.return_value),
                mock.call().download_as_bytes(),
                mock.call("gs://bucket/remote/log/location/1.log", mock_client.return_value),
                mock.call().upload_from_string(
                    "MESSAGE\n",
                    content_type="text/plain",
                ),
            ],
            any_order=False,
        )

    @pytest.mark.parametrize(
        ("delete_local_copy", "expected_existence_of_local_copy"),
        [(True, False), (False, True)],
    )
    @mock.patch(
        "airflow.providers.google.cloud.log.gcs_task_handler.get_credentials_and_project_id",
        return_value=("TEST_CREDENTIALS", "TEST_PROJECT_ID"),
    )
    @mock.patch("google.cloud.storage.Client")
    @mock.patch("google.cloud.storage.Blob")
    def test_close_with_delete_local_copy_conf(
        self,
        mock_blob,
        mock_client,
        mock_creds,
        local_log_location,
        delete_local_copy,
        expected_existence_of_local_copy,
    ):
        mock_blob.from_string.return_value.open.return_value = io.TextIOWrapper(
            io.BytesIO(b"CONTENT"), encoding="utf-8"
        )
        with conf_vars({("logging", "delete_local_logs"): str(delete_local_copy)}):
            handler = GCSTaskHandler(
                base_log_folder=local_log_location,
                gcs_log_folder="gs://bucket/remote/log/location",
            )

        handler.log.info("test")
        handler.set_context(self.ti)
        assert handler.upload_on_close

        handler.close()
        assert os.path.exists(handler.handler.baseFilename) == expected_existence_of_local_copy

    @pytest.fixture(autouse=True)
    def test_filename_template_for_backward_compatibility(self, local_log_location):
        # filename_template arg support for running the latest provider on airflow 2
        GCSTaskHandler(
            base_log_folder=local_log_location,
            gcs_log_folder="gs://bucket/remote/log/location",
            filename_template=None,
        )
